# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from zun.conf import path

image_driver_opts = [
    cfg.ListOpt(
        'image_driver_list',
        default=['glance', 'docker'],
        help="""Defines the list of image driver to use for downloading image.
Possible values:
* ``docker``
* ``glance``
Services which consume this:
* ``zun-compute``
Interdependencies to other options:
* None
"""),
    cfg.StrOpt(
        'default_image_driver',
        default='docker',
        help='The default container image driver to use.'),
]

sandbox_opts = [
    cfg.StrOpt(
        'sandbox_image',
        default='kubernetes/pause',
        help='Container image for sandbox container.'),
    cfg.StrOpt(
        'sandbox_image_driver',
        default='docker',
        help='Image driver for sandbox container.'),
    cfg.StrOpt(
        'sandbox_image_pull_policy',
        default='ifnotpresent',
        help='Image pull policy for sandbox image.'),
]

glance_driver_opts = [
    cfg.StrOpt(
        'images_directory',
        default=path.state_path_def('images'),
        help='Shared directory where glance images located. If '
             'specified, docker will try to load the image from '
             'the shared directory by image ID.'),
]

glance_opt_group = cfg.OptGroup(name='glance',
                                title='Glance options for image management')

DEFAULT_OPTS = (image_driver_opts + sandbox_opts)
GLANCE_OPTS = (glance_driver_opts)


def register_opts(conf):
    conf.register_group(glance_opt_group)
    conf.register_opts(glance_driver_opts, group=glance_opt_group)
    conf.register_opts(image_driver_opts)
    conf.register_opts(sandbox_opts)


def list_opts():
    return {"DEFAULT": DEFAULT_OPTS, glance_opt_group: GLANCE_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add disk to container

Revision ID: d9714eadbdc2
Revises: 71f8b4cf1dbf
Create Date: 2018-01-30 13:47:11.871600

"""

# revision identifiers, used by Alembic.
revision = 'd9714eadbdc2'
down_revision = '71f8b4cf1dbf'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    with op.batch_alter_table('container', schema=None) as batch_op:
        batch_op.add_column(sa.Column('disk', sa.Integer(), nullable=True))



# Copyright 2010-2011 OpenStack Foundation
# Copyright 2012-2013 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for database migrations. There are "opportunistic" tests for both mysql
and postgresql in here, which allows testing against these databases in a
properly configured unit test environment.

For the opportunistic testing you need to set up a db named 'openstack_citest'
with user 'openstack_citest' and password 'openstack_citest' on localhost.
The test will then use that db and u/p combo to run the tests.

For postgres on Ubuntu this can be done with the following commands:

::

 sudo -u postgres psql
 postgres=# create user openstack_citest with createdb login password
      'openstack_citest';
 postgres=# create database openstack_citest with owner openstack_citest;

"""

import contextlib
import os
from unittest import mock

from alembic import script
import fixtures
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_base
from oslo_log import log as logging
import sqlalchemy
import sqlalchemy.exc

from zun.db.sqlalchemy import migration
from zun.tests import base


LOG = logging.getLogger(__name__)

# NOTE(vdrok): This was introduced after migration tests started taking more
# time in gate. Timeout value in seconds for tests performing migrations.
MIGRATIONS_TIMEOUT = 300


def _get_connect_string(backend, user, passwd, database):
    """Get database connection

    Try to get a connection with a very specific set of values, if we get
    these then we'll run the tests, otherwise they are skipped
    """
    if backend == "postgres":
        backend = "postgresql+psycopg2"
    elif backend == "mysql":
        backend = "mysql+pymysql"
    elif backend == "sqlite":
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))
        return 'sqlite:///' + os.path.join(path, database)
    else:
        raise Exception("Unrecognized backend: '%s'" % backend)

    return ("%(backend)s://%(user)s:%(passwd)s@localhost/%(database)s"
            % {'backend': backend, 'user': user, 'passwd': passwd,
               'database': database})


def _is_backend_avail(backend, user, passwd, database):
    try:
        connect_uri = _get_connect_string(backend, user, passwd, database)
        engine = sqlalchemy.create_engine(connect_uri)
        connection = engine.connect()
    except Exception:
        # intentionally catch all to handle exceptions even if we don't
        # have any backend code loaded.
        return False
    else:
        connection.close()
        engine.dispose()
        return True


@contextlib.contextmanager
def patch_with_engine(engine):
    with mock.patch.object(enginefacade.get_legacy_facade(),
                           'get_engine') as patch_engine:
        patch_engine.return_value = engine
        yield


class WalkVersionsMixin(object):
    def _walk_versions(self, engine=None, alembic_cfg=None):
        # Determine latest version script from the repo, then
        # upgrade from 1 through to the latest, with no data
        # in the databases. This just checks that the schema itself
        # upgrades successfully.
        # Place the database under version control
        with patch_with_engine(engine):

            script_directory = script.ScriptDirectory.from_config(alembic_cfg)
            self.assertIsNone(self.migration_api.version(alembic_cfg))

            versions = [ver for ver in script_directory.walk_revisions()]

            for version in reversed(versions):
                self._migrate_up(engine, alembic_cfg,
                                 version.revision, with_data=True)

    def _migrate_up(self, engine, config, version, with_data=False):
        """migrate up to a new version of the db.

        We allow for data insertion and post checks at every
        migration version with special _pre_upgrade_### and
        _check_### functions in the main test.
        """
        # NOTE(sdague): try block is here because it's impossible to debug
        # where a failed data migration happens otherwise
        try:
            if with_data:
                data = None
                pre_upgrade = getattr(
                    self, "_pre_upgrade_%s" % version, None)
                if pre_upgrade:
                    data = pre_upgrade(engine)

            self.migration_api.upgrade(version)
            self.assertEqual(version, self.migration_api.version(config))
            if with_data:
                check = getattr(self, "_check_%s" % version, None)
                if check:
                    check(engine, data)
        except Exception:
            LOG.error("Failed to migrate to version %(version)s on engine "
                      "%(engine)s",
                      {'version': version, 'engine': engine})
            raise


class TestWalkVersions(base.TestCase, WalkVersionsMixin):
    def setUp(self):
        super(TestWalkVersions, self).setUp()
        self.migration_api = mock.MagicMock()
        self.engine = mock.MagicMock()
        self.config = mock.MagicMock()
        self.versions = [mock.Mock(revision='2b2'), mock.Mock(revision='1a1')]

    def test_migrate_up(self):
        self.migration_api.version.return_value = 'dsa123'

        self._migrate_up(self.engine, self.config, 'dsa123')

        self.migration_api.upgrade.assert_called_with('dsa123')
        self.migration_api.version.assert_called_with(self.config)

    def test_migrate_up_with_data(self):
        test_value = {"a": 1, "b": 2}
        self.migration_api.version.return_value = '141'
        self._pre_upgrade_141 = mock.MagicMock()
        self._pre_upgrade_141.return_value = test_value
        self._check_141 = mock.MagicMock()

        self._migrate_up(self.engine, self.config, '141', True)

        self._pre_upgrade_141.assert_called_with(self.engine)
        self._check_141.assert_called_with(self.engine, test_value)

    @mock.patch.object(script, 'ScriptDirectory')
    @mock.patch.object(WalkVersionsMixin, '_migrate_up')
    def test_walk_versions_all_default(self, _migrate_up, script_directory):
        fc = script_directory.from_config()
        fc.walk_revisions.return_value = self.versions
        self.migration_api.version.return_value = None

        self._walk_versions(self.engine, self.config)

        self.migration_api.version.assert_called_with(self.config)

        upgraded = [mock.call(self.engine, self.config, v.revision,
                    with_data=True) for v in reversed(self.versions)]
        self.assertEqual(self._migrate_up.call_args_list, upgraded)

    @mock.patch.object(script, 'ScriptDirectory')
    @mock.patch.object(WalkVersionsMixin, '_migrate_up')
    def test_walk_versions_all_false(self, _migrate_up, script_directory):
        fc = script_directory.from_config()
        fc.walk_revisions.return_value = self.versions
        self.migration_api.version.return_value = None

        self._walk_versions(self.engine, self.config)

        upgraded = [mock.call(self.engine, self.config, v.revision,
                    with_data=True) for v in reversed(self.versions)]
        self.assertEqual(upgraded, self._migrate_up.call_args_list)


class MigrationCheckersMixin(object):
    def setUp(self):
        super(MigrationCheckersMixin, self).setUp()

        self.config = migration._alembic_config()
        self.migration_api = migration
        self.useFixture(fixtures.Timeout(MIGRATIONS_TIMEOUT, gentle=True))

    def test_walk_versions(self):
        db_name = str(self.engine.url).rsplit('/', 1)[-1]
        self.config.attributes['test_db_name'] = db_name
        self._walk_versions(self.engine, self.config)

    def test_connect_fail(self):
        """Test that we can trigger a database connection failure

        Test that we can fail gracefully to ensure we don't break people
        without specific database backend.

        Pass on Sqlite since it has no concept of connecting as a user
        """
        if _is_backend_avail(self.FIXTURE.DRIVER, "zun_migration_fail",
                             self.FIXTURE.USERNAME, self.FIXTURE.DBNAME):
            if self.FIXTURE.DRIVER != 'sqlite':
                self.fail("Shouldn't have connected")

    def test_upgrade_and_version(self):
        with patch_with_engine(self.engine):
            self.migration_api.upgrade('head')
            self.assertIsNotNone(self.migration_api.version())

    def test_create_schema_and_version(self):
        with patch_with_engine(self.engine):
            self.migration_api.create_schema()
            self.assertIsNotNone(self.migration_api.version())

    # TODO(swatson): Find a patch version to make the next two tests make sense
    def test_upgrade_and_create_schema(self):
        with patch_with_engine(self.engine):
            self.migration_api.upgrade('2d1354bbf76e')
            self.assertRaises(db_exc.DBMigrationError,
                              self.migration_api.create_schema)

    def test_upgrade_twice(self):
        with patch_with_engine(self.engine):
            self.migration_api.upgrade('2d1354bbf76e')
            v1 = self.migration_api.version()
            self.migration_api.upgrade('head')
            v2 = self.migration_api.version()
            self.assertNotEqual(v1, v2)


class TestMigrationsMySQL(MigrationCheckersMixin,
                          WalkVersionsMixin,
                          test_base.MySQLOpportunisticTestCase):
    def _setUp(self):
        self.addCleanup(self.engine.dispose)


# Investigate if there is model/migration sync issue in zun and enable testing.
# See implementation in ironic's test_migrations.ModelsMigrationsSyncMysql for
# re-enabling sync tests.

# TODO(swatson): Enable SQLite tests once SQLite migration issues are fixed
# SQLite doesn't support "drop column" or "alter table" calls. We have at least
# 2 migrations using drop columns, and at least 1 using an alter table call.
# Might need to modify those migrations for SQLite compatibility.



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pecan

from zun.api.controllers import base
from zun.api.controllers.v1 import collection
from zun.api.controllers.v1.views import hosts_view as view
from zun.api import utils as api_utils
from zun.common import exception
from zun.common import policy
from zun import objects


def check_policy_on_host(host, action):
    context = pecan.request.context
    policy.enforce(context, action, host, action=action)


class HostCollection(collection.Collection):
    """API representation of a collection of hosts."""

    fields = {
        'hosts',
        'next'
    }

    """A list containing compute node objects"""

    def __init__(self, **kwargs):
        super(HostCollection, self).__init__(**kwargs)
        self._type = 'hosts'

    @staticmethod
    def convert_with_links(nodes, limit, url=None,
                           expand=False, **kwargs):
        collection = HostCollection()
        collection.hosts = [view.format_host(url, p) for p in nodes]
        collection.next = collection.get_next(limit, url=url, **kwargs)
        return collection


class HostController(base.Controller):
    """Host info controller"""

    @pecan.expose('json')
    @base.Controller.api_version("1.4")
    @exception.wrap_pecan_controller_exception
    def get_all(self, **kwargs):
        """Retrieve a list of hosts"""
        context = pecan.request.context
        policy.enforce(context, "host:get_all",
                       action="host:get_all")
        return self._get_host_collection(**kwargs)

    def _get_host_collection(self, **kwargs):
        context = pecan.request.context
        limit = api_utils.validate_limit(kwargs.get('limit'))
        sort_dir = api_utils.validate_sort_dir(kwargs.get('sort_dir', 'asc'))
        sort_key = kwargs.get('sort_key', 'hostname')
        expand = kwargs.get('expand')
        filters = None
        marker_obj = None
        resource_url = kwargs.get('resource_url')
        marker = kwargs.get('marker')
        if marker:
            marker_obj = objects.ComputeNode.get_by_uuid(context, marker)
        nodes = objects.ComputeNode.list(context,
                                         limit,
                                         marker_obj,
                                         sort_key,
                                         sort_dir,
                                         filters=filters)
        return HostCollection.convert_with_links(nodes, limit,
                                                 url=resource_url,
                                                 expand=expand,
                                                 sort_key=sort_key,
                                                 sort_dir=sort_dir)

    @pecan.expose('json')
    @base.Controller.api_version("1.4")
    @exception.wrap_pecan_controller_exception
    def get_one(self, host_ident):
        """Retrieve information about the given host.

        :param host_ident: UUID or name of a host.
        """
        context = pecan.request.context
        policy.enforce(context, "host:get", action="host:get")
        host = api_utils.get_resource('ComputeNode', host_ident)
        return view.format_host(pecan.request.host_url, host)



#    Copyright 2016 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from zun.common import exception
from zun.compute import rpcapi
from zun import objects
from zun.tests import base
from zun.tests.unit.db import utils


class TestAPI(base.TestCase):

    def setUp(self):
        super(TestAPI, self).setUp()
        self.compute_rpcapi = rpcapi.API()

    @mock.patch('zun.api.servicegroup.ServiceGroup.service_is_up')
    @mock.patch('zun.objects.ZunService.list_by_binary')
    @mock.patch('zun.common.rpc_service.API._call')
    def test_container_delete_with_host_no_tup(self, mock_rpc_call,
                                               mock_list, mock_service_is_up):
        test_container = utils.get_test_container()
        test_container_obj = objects.Container(self.context, **test_container)
        test_service = utils.get_test_zun_service(host="fake_host")
        test_service_obj = objects.ZunService(self.context, **test_service)
        mock_list.return_value = [test_service_obj]
        mock_service_is_up.return_value = False
        self.assertRaises(exception.ContainerHostNotUp,
                          self.compute_rpcapi.container_delete,
                          self.context, test_container_obj, False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_vif.objects import vif as obj_osvif
from oslo_versionedobjects import fields as obj_fields

from zun.common import consts
from zun.objects import base
from zun.objects import fields as zun_fields


@base.ZunObjectRegistry.register
class VIFState(base.ZunObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    # FIXME(dulek): I know it's an ugly hack, but turns out you cannot
    #               serialize-deserialize objects containing objects from
    #               different namespaces, so we need 'os_vif' namespace here.
    OBJ_PROJECT_NAMESPACE = 'os_vif'
    OBJ_SERIAL_NAMESPACE = 'versioned_object'

    fields = {
        'default_vif': obj_fields.ObjectField(obj_osvif.VIFBase.__name__,
                                              subclasses=True, nullable=False),
        'additional_vifs': zun_fields.DictOfVIFsField(default={}),
    }

    @property
    def vifs(self):
        d = {
            consts.DEFAULT_IFNAME: self.default_vif,
        }
        if self.obj_attr_is_set('additional_vifs'):
            d.update(self.additional_vifs)
        return d



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_vif.objects import vif
from oslo_serialization import jsonutils as json
from oslo_versionedobjects import fields

from zun.common import consts

UnspecifiedDefault = fields.UnspecifiedDefault


class BaseZunEnum(fields.Enum):
    def __init__(self, **kwargs):
        super(BaseZunEnum, self).__init__(valid_values=self.__class__.ALL)


class ContainerStatus(fields.Enum):
    ALL = consts.CONTAINER_STATUSES

    def __init__(self):
        super(ContainerStatus, self).__init__(
            valid_values=ContainerStatus.ALL)


class ContainerStatusField(fields.BaseEnumField):
    AUTO_TYPE = ContainerStatus()


class CapsuleStatus(fields.Enum):
    ALL = consts.CAPSULE_STATUSES

    def __init__(self):
        super(CapsuleStatus, self).__init__(
            valid_values=CapsuleStatus.ALL)


class CapsuleStatusField(fields.BaseEnumField):
    AUTO_TYPE = CapsuleStatus()


class TaskState(fields.Enum):
    ALL = consts.TASK_STATES

    def __init__(self):
        super(TaskState, self).__init__(
            valid_values=TaskState.ALL)


class TaskStateField(fields.BaseEnumField):
    AUTO_TYPE = TaskState()


class ListOfIntegersField(fields.AutoTypedField):
    AUTO_TYPE = fields.List(fields.Integer())


class Json(fields.FieldType):
    def coerce(self, obj, attr, value):
        if isinstance(value, str):
            loaded = json.loads(value)
            return loaded
        return value

    def from_primitive(self, obj, attr, value):
        return self.coerce(obj, attr, value)

    def to_primitive(self, obj, attr, value):
        return json.dumps(value)


class JsonField(fields.AutoTypedField):
    AUTO_TYPE = Json()


class SetOfStringsField(fields.AutoTypedField):
    AUTO_TYPE = fields.Set(fields.String())


class ListOfListsOfStringsField(fields.AutoTypedField):
    AUTO_TYPE = fields.List(fields.List(fields.String()))


class ResourceClass(fields.Enum):
    ALL = consts.RESOURCE_CLASSES

    def __init__(self):
        super(ResourceClass, self).__init__(
            valid_values=ResourceClass.ALL)


class ResourceClassField(fields.AutoTypedField):
    AUTO_TYPE = ResourceClass()


class PciDeviceStatus(BaseZunEnum):

    AVAILABLE = "available"
    CLAIMED = "claimed"
    ALLOCATED = "allocated"
    REMOVED = "removed"  # The device has been hot-removed and not yet deleted
    DELETED = "deleted"  # The device is marked not available/deleted.
    UNCLAIMABLE = "unclaimable"
    UNAVAILABLE = "unavailable"

    ALL = (AVAILABLE, CLAIMED, ALLOCATED, REMOVED, DELETED, UNAVAILABLE,
           UNCLAIMABLE)


class PciDeviceType(BaseZunEnum):

    STANDARD = "PCI"
    SRIOV_PF = "PF"
    SRIOV_VF = "VF"

    ALL = (STANDARD, SRIOV_PF, SRIOV_VF)


class PciDeviceTypeField(fields.BaseEnumField):
    AUTO_TYPE = PciDeviceType()


class PciDeviceStatusField(fields.BaseEnumField):
    AUTO_TYPE = PciDeviceStatus()


class DictOfVIFsField(fields.AutoTypedField):
    AUTO_TYPE = fields.Dict(fields.Object(vif.VIFBase.__name__,
                                          subclasses=True))



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from zun.scheduler import filters

LOG = logging.getLogger(__name__)


class DiskFilter(filters.BaseHostFilter):
    """DEPRECATED: Filter the containers by disk request"""

    run_filter_once_per_request = True

    def __init__(self):
        super(DiskFilter, self).__init__()
        LOG.warning('The DiskFilter is deprecated since the 7.0.0 '
                    'Wallaby release. DISK_GB filtering is performed natively '
                    'using the Placement service when using the '
                    'filter_scheduler driver. Operators should define disk '
                    'allocation ratios either per host in the zun.conf '
                    'or via the placement API.')

    def host_passes(self, host_state, container, extra_spec):
        if not hasattr(container, 'disk') or not container.disk:
            return True

        if not host_state.disk_quota_supported:
            LOG.debug("(%(host_state)s) does not support disk quota, but the "
                      "container requires disk quota of %(container_disk)d.",
                      {'host_state': host_state,
                       'container_disk': container.disk})
            return False

        usable_disk = host_state.disk_total - host_state.disk_used
        if usable_disk < container.disk:
            LOG.debug("%(host_state)s does not have %(container_disk)d "
                      "usable disk, it only has %(usable_disk)d usable "
                      "disk.",
                      {'host_state': host_state,
                       'container_disk': container.disk,
                       'usable_disk': usable_disk})
            return False
        host_state.limits['disk'] = host_state.disk_total
        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from zun.api.controllers.v1.schemas import parameter_types

_legacy_container_properties = {
    'name': parameter_types.container_name,
    'image': parameter_types.image_name,
    'command': parameter_types.command,
    'cpu': parameter_types.cpu,
    'cpu_policy': parameter_types.cpu_policy,
    'memory': parameter_types.memory,
    'workdir': parameter_types.workdir,
    'auto_remove': parameter_types.auto_remove,
    'image_pull_policy': parameter_types.image_pull_policy,
    'labels': parameter_types.labels,
    'environment': parameter_types.environment,
    'restart_policy': parameter_types.restart_policy,
    'interactive': parameter_types.boolean,
    'tty': parameter_types.boolean,
    'image_driver': parameter_types.image_driver,
    'security_groups': parameter_types.security_groups,
    'hints': parameter_types.hints,
    'mounts': parameter_types.mounts,
    'nets': parameter_types.nets,
    'runtime': parameter_types.runtime,
    'hostname': parameter_types.hostname,
    'disk': parameter_types.disk,
    'availability_zone': parameter_types.availability_zone,
    'auto_heal': parameter_types.boolean,
    'privileged': parameter_types.boolean,
    'healthcheck': parameter_types.healthcheck,
    'exposed_ports': parameter_types.exposed_ports,
    'registry': parameter_types.container_registry,
}

legacy_container_create = {
    'type': 'object',
    'properties': _legacy_container_properties,
    'required': ['image'],
    'additionalProperties': False
}

_container_properties = copy.deepcopy(_legacy_container_properties)
_container_properties['command'] = parameter_types.command_list

container_create = {
    'type': 'object',
    'properties': _container_properties,
    'allOf': [
        {
            'required': ['image'],
        },
        {
            'not': {'required': ['security_groups', 'exposed_ports']}
        }
    ],
    'additionalProperties': False
}

# Add host in container
container_create_v139 = copy.deepcopy(container_create)
container_create_v139['properties']['host'] = parameter_types.hostname

# Add entrypoint in container
container_create_v140 = copy.deepcopy(container_create_v139)
container_create_v140['properties']['entrypoint'] = parameter_types.entrypoint

query_param_rename = {
    'type': 'object',
    'properties': {
        'name': parameter_types.container_name
    },
    'additionalProperties': False
}

query_param_create = {
    'type': 'object',
    'properties': {
        'run': parameter_types.boolean_extended
    },
    'additionalProperties': False
}

_container_update_properties = {
    'cpu': parameter_types.cpu,
    'memory': parameter_types.memory,
    'name': parameter_types.container_name,
    'auto_heal': parameter_types.boolean
}

container_update = {
    'type': 'object',
    'properties': _container_update_properties,
    'additionalProperties': False
}

query_param_delete = {
    'type': 'object',
    'properties': {
        'force': parameter_types.boolean_extended,
        'all_projects': parameter_types.boolean_extended,
        'stop': parameter_types.boolean_extended
    },
    'additionalProperties': False
}

query_param_reboot = {
    'type': 'object',
    'properties': {
        'timeout': parameter_types.non_negative_integer
    },
    'additionalProperties': False
}

query_param_logs = {
    'type': 'object',
    'properties': {
        'stdout': parameter_types.boolean_extended,
        'stderr': parameter_types.boolean_extended,
        'timestamps': parameter_types.boolean_extended,
        'tail': parameter_types.str_and_int,
        'since': parameter_types.logs_since
    },
    'additionalProperties': False
}

query_param_top = {
    'type': 'object',
    'properties': {
        'ps_args': parameter_types.string_ps_args
    },
    'additionalProperties': False
}
query_param_stop = copy.deepcopy(query_param_reboot)

query_param_resize = {
    'type': 'object',
    'properties': {
        'h': parameter_types.positive_integer,
        'w': parameter_types.positive_integer
    },
    'additionalProperties': False
}

query_param_execute_resize = {
    'type': 'object',
    'properties': {
        'exec_id': parameter_types.exec_id,
        'h': parameter_types.positive_integer,
        'w': parameter_types.positive_integer
    },
    'additionalProperties': False
}

query_param_signal = {
    'type': 'object',
    'properties': {
        'signal': parameter_types.signal
    },
    'additionalProperties': False
}

query_param_execute_command = {
    'type': 'object',
    'properties': {
        'run': parameter_types.boolean,
        'interactive': parameter_types.boolean,
        'command': parameter_types.exec_command,
    },
    'additionalProperties': False
}

query_param_commit = {
    'type': 'object',
    'properties': {
        'repository': parameter_types.string_ps_args,
        'tag': parameter_types.string_ps_args
    },
    'required': ['repository'],
    'additionalProperties': False
}
add_security_group = {
    'type': 'object',
    'properties': {
        'name': {
            'type': 'string',
            'minLength': 1,
            'maxLength': 255
        }
    },
    'additionalProperties': False
}

remove_security_group = copy.deepcopy(add_security_group)

network_detach = {
    'type': 'object',
    'properties': {
        'network': {
            'type': 'string',
            'minLength': 1,
            'maxLength': 255,
        },
        'port': {
            'type': 'string',
            'minLength': 1,
            'maxLength': 255,
        }
    },
    'oneOf': [
        {
            'required': ['network']
        },
        {
            'required': ['port']
        }
    ],
    'additionalProperties': False
}

network_attach = {
    'type': 'object',
    'properties': {
        'network': {
            'type': ['string'],
            'minLength': 1,
            'maxLength': 255,
        },
        'fixed_ip': {
            'type': ['string'],
            'oneOf': [
                {'format': 'ipv4'},
                {'format': 'ipv6'}
            ]
        },
        'port': {
            'type': ['string'],
            'maxLength': 255,
            'minLength': 1,
        }
    },
    'additionalProperties': False,
    'not': {'required': ['port', 'network']}
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from kuryr.lib.binding.drivers import utils as kl_utils
from kuryr.lib import constants as kl_const
from os_vif.objects import fixed_ip as osv_fixed_ip
from os_vif.objects import network as osv_network
from os_vif.objects import route as osv_route
from os_vif.objects import subnet as osv_subnet
from os_vif.objects import vif as osv_vif
from oslo_config import cfg as oslo_cfg
from stevedore import driver as stv_driver

from zun.common import exception
from zun.common import utils


# REVISIT(ivc): consider making this module part of kuryr-lib
_VIF_TRANSLATOR_NAMESPACE = "zun.vif_translators"
_VIF_MANAGERS = {}


def _neutron_to_osvif_network(os_network):
    """Converts Neutron network to os-vif Subnet.

    :param os_network: openstack.network.v2.netwrork.Network object.
    :return: an os-vif Network object
    """

    obj = osv_network.Network(id=os_network['id'])

    if os_network['name'] is not None:
        obj.label = os_network['name']

    if os_network['mtu'] is not None:
        obj.mtu = os_network['mtu']

    # Vlan information will be used later in Sriov binding driver
    if os_network['provider:network_type'] == 'vlan':
        obj.should_provide_vlan = True
        obj.vlan = os_network['provider:segmentation_id']

    return obj


def _neutron_to_osvif_subnet(os_subnet):
    """Converts Neutron subnet to os-vif Subnet.

    :param os_subnet: openstack.network.v2.subnet.Subnet object
    :return: an os-vif Subnet object
    """

    obj = osv_subnet.Subnet(
        cidr=os_subnet['cidr'],
        dns=os_subnet['dns_nameservers'],
        routes=_neutron_to_osvif_routes(os_subnet['host_routes']))

    if os_subnet['gateway_ip'] is not None:
        obj.gateway = os_subnet['gateway_ip']

    return obj


def _neutron_to_osvif_routes(neutron_routes):
    """Converts Neutron host_routes to os-vif RouteList.

    :param neutron_routes: list of routes as returned by neutron client's
                           'show_subnet' in 'host_routes' attribute
    :return: an os-vif RouteList object
    """

    # NOTE(gryf): Nested attributes for OpenStackSDK objects are simple types,
    # like dicts and lists, that's why neutron_routes is a list of dicts.
    obj_list = [osv_route.Route(cidr=route['destination'],
                                gateway=route['nexthop'])
                for route in neutron_routes]

    return osv_route.RouteList(objects=obj_list)


def _make_vif_subnet(subnets, subnet_id):
    """Makes a copy of an os-vif Subnet from subnets mapping.

    :param subnets: subnet mapping
    :param subnet_id: ID of the subnet to extract from 'subnets' mapping
    :return: a copy of an os-vif Subnet object matching 'subnet_id'
    """
    subnet = subnets[subnet_id]

    vif_subnet = _neutron_to_osvif_subnet(subnet)
    vif_subnet.ips = osv_fixed_ip.FixedIPList(objects=[])
    return vif_subnet


def _make_vif_subnets(neutron_port, subnets):
    """Gets a list of os-vif Subnet objects for port.

    :param neutron_port: dict containing port information as returned by
                         neutron client's 'show_port'
    :param subnets: subnet mapping
    :return: list of os-vif Subnet object
    """

    vif_subnets = {}

    for neutron_fixed_ip in neutron_port.get('fixed_ips', []):
        subnet_id = neutron_fixed_ip['subnet_id']
        ip_address = neutron_fixed_ip['ip_address']

        if subnet_id not in subnets:
            continue

        try:
            subnet = vif_subnets[subnet_id]
        except KeyError:
            subnet = _make_vif_subnet(subnets, subnet_id)
            vif_subnets[subnet_id] = subnet

        subnet.ips.objects.append(osv_fixed_ip.FixedIP(address=ip_address))

    if not vif_subnets:
        raise exception.ZunException(
            "No valid subnets found for port %(port_id)s"
            % {'port_id': neutron_port.get('id')})

    return list(vif_subnets.values())


def _make_vif_network(neutron_port, network, subnets):
    """Get an os-vif Network object for port.

    :param neutron_port: dict containing port information as returned by
                         neutron client's 'show_port'
    :param subnets: subnet mapping
    :return: os-vif Network object
    """

    vif_network = _neutron_to_osvif_network(network)
    vif_network.subnets = osv_subnet.SubnetList(
        objects=_make_vif_subnets(neutron_port, subnets))

    return vif_network


def _get_vif_name(neutron_port):
    """Gets a VIF device name for port.

    :param neutron_port: dict containing port information as returned by
                         neutron client's 'show_port'
    """

    vif_name, _ = kl_utils.get_veth_pair_names(neutron_port['id'])
    return vif_name


def _get_ovs_hybrid_bridge_name(neutron_port):
    """Gets a name of the Linux bridge name for hybrid OpenVSwitch port.

    :param neutron_port: dict containing port information as returned by
                         neutron client's 'show_port'
    """
    return ('qbr' + neutron_port['id'])[:kl_const.NIC_NAME_LEN]


def neutron_to_osvif_vif_ovs(vif_plugin, neutron_port, network, subnets):
    """Converts Neutron port to VIF object for os-vif 'ovs' plugin.

    :param vif_plugin: name of the os-vif plugin to use (i.e. 'ovs')
    :param neutron_port: dict containing port information as returned by
                         neutron client's 'show_port'
    :param subnets: subnet mapping
    :return: os-vif VIF object
    """

    profile = osv_vif.VIFPortProfileOpenVSwitch(
        interface_id=neutron_port['id'])

    details = neutron_port.get('binding:vif_details', {})
    ovs_bridge = details.get('bridge_name',
                             oslo_cfg.CONF.neutron.ovs_bridge)
    if not ovs_bridge:
        raise oslo_cfg.RequiredOptError('ovs_bridge', 'neutron_defaults')

    network = _make_vif_network(neutron_port, network, subnets)
    network.bridge = ovs_bridge

    if details.get('ovs_hybrid_plug'):
        vif = osv_vif.VIFBridge(
            id=neutron_port['id'],
            address=neutron_port['mac_address'],
            network=network,
            has_traffic_filtering=details.get('port_filter', False),
            preserve_on_delete=False,
            active=utils.is_port_active(neutron_port),
            port_profile=profile,
            plugin=vif_plugin,
            vif_name=_get_vif_name(neutron_port),
            bridge_name=_get_ovs_hybrid_bridge_name(neutron_port))
    else:
        vif = osv_vif.VIFOpenVSwitch(
            id=neutron_port['id'],
            address=neutron_port['mac_address'],
            network=network,
            has_traffic_filtering=details.get('port_filter', False),
            preserve_on_delete=False,
            active=utils.is_port_active(neutron_port),
            port_profile=profile,
            plugin=vif_plugin,
            vif_name=_get_vif_name(neutron_port),
            bridge_name=network.bridge)

    return vif


def neutron_to_osvif_vif(vif_translator, neutron_port, network, subnets):
    """Converts Neutron port to os-vif VIF object.

    :param vif_translator: name of the traslator for the os-vif plugin to use
    :param neutron_port: dict containing port information as returned by
                         neutron client
    :param subnets: subnet mapping
    :return: os-vif VIF object
    """

    try:
        mgr = _VIF_MANAGERS[vif_translator]
    except KeyError:
        mgr = stv_driver.DriverManager(
            namespace=_VIF_TRANSLATOR_NAMESPACE,
            name=vif_translator, invoke_on_load=False)
        _VIF_MANAGERS[vif_translator] = mgr

    return mgr.driver(vif_translator, neutron_port, network, subnets)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64

from cryptography import fernet
from oslo_config import cfg
from oslo_utils import encodeutils

from zun.common import exception


def encrypt(value, encryption_key=None):
    if value is None:
        return None

    encryption_key = get_valid_encryption_key(encryption_key)
    encoded_key = base64.b64encode(encryption_key.encode('utf-8'))
    sym = fernet.Fernet(encoded_key)
    res = sym.encrypt(encodeutils.safe_encode(value))
    return encodeutils.safe_decode(res)


def decrypt(data, encryption_key=None):
    if data is None:
        return None

    encryption_key = get_valid_encryption_key(encryption_key)
    encoded_key = base64.b64encode(encryption_key.encode('utf-8'))
    sym = fernet.Fernet(encoded_key)
    try:
        value = sym.decrypt(encodeutils.safe_encode(data))
        if value is not None:
            return encodeutils.safe_decode(value, 'utf-8')
    except fernet.InvalidToken:
        raise exception.InvalidEncryptionKey()


def get_valid_encryption_key(encryption_key):
    if encryption_key is None:
        encryption_key = cfg.CONF.auth_encryption_key

    return encryption_key[:32]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent import futures
import multiprocessing
import os
import threading
import time

import cotyledon
import flask
from futurist import periodics
from http import client as httplib
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from pyroute2.ipdb import transactional

from zun.cni.plugins import zun_cni_registry
from zun.cni import utils as cni_utils
from zun.common import context as zun_context
from zun.common import exception
from zun.common import utils
from zun.network import neutron
from zun import objects

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class DaemonServer(object):
    def __init__(self, plugin):
        self.plugin = plugin
        self.application = flask.Flask('zun-cni-daemon')
        self.application.add_url_rule(
            '/addNetwork', methods=['POST'], view_func=self.add)
        self.application.add_url_rule(
            '/delNetwork', methods=['POST'], view_func=self.delete)
        self.headers = {'ContentType': 'application/json',
                        'Connection': 'close'}

    def _prepare_request(self):
        params = cni_utils.CNIParameters(flask.request.get_json())
        LOG.debug('Received %s request. CNI Params: %s',
                  params.CNI_COMMAND, params)
        return params

    def add(self):
        try:
            params = self._prepare_request()
        except Exception:
            LOG.exception('Exception when reading CNI params.')
            return '', httplib.BAD_REQUEST, self.headers

        try:
            vif = self.plugin.add(params)
            data = jsonutils.dumps(vif.obj_to_primitive())
        except exception.ResourceNotReady:
            LOG.error('Error when processing addNetwork request')
            return '', httplib.GATEWAY_TIMEOUT, self.headers
        except Exception:
            LOG.exception('Error when processing addNetwork request. CNI '
                          'Params: %s', params)
            return '', httplib.INTERNAL_SERVER_ERROR, self.headers

        return data, httplib.ACCEPTED, self.headers

    def delete(self):
        try:
            params = self._prepare_request()
        except Exception:
            LOG.exception('Exception when reading CNI params.')
            return '', httplib.BAD_REQUEST, self.headers

        try:
            self.plugin.delete(params)
        except exception.ResourceNotReady:
            # NOTE(dulek): It's better to ignore this error - most of the time
            # it will happen when capsule/container is long gone and runtime
            # overzealously tries to delete it from the network. We cannot
            # really do anything without VIF metadata, so let's just tell
            # runtime to move along.
            LOG.warning('Error when processing delNetwork request. '
                        'Ignoring this error, capsule/container is most '
                        'likely gone')
            return '', httplib.NO_CONTENT, self.headers
        except Exception:
            LOG.exception('Error when processing delNetwork request. CNI '
                          'Params: %s.', params)
            return '', httplib.INTERNAL_SERVER_ERROR, self.headers
        return '', httplib.NO_CONTENT, self.headers

    def run(self):
        address = CONF.cni_daemon.cni_daemon_host
        port = CONF.cni_daemon.cni_daemon_port

        try:
            self.application.run(address, port, threaded=False,
                                 processes=CONF.cni_daemon.worker_num)
        except Exception:
            LOG.exception('Failed to start zun-cni-daemon.')
            raise


class CNIDaemonServerService(cotyledon.Service):
    name = "server"

    def __init__(self, worker_id, registry):
        super(CNIDaemonServerService, self).__init__(worker_id)
        self.registry = registry
        self.plugin = zun_cni_registry.ZunCNIRegistryPlugin(registry)
        self.server = DaemonServer(self.plugin)

    def run(self):
        # NOTE(dulek): We might do a *lot* of pyroute2 operations, let's
        #              make the pyroute2 timeout configurable to make sure
        #              kernel will have chance to catch up.
        transactional.SYNC_TIMEOUT = CONF.cni_daemon.pyroute2_timeout

        # Run HTTP server
        self.server.run()


class CNIDaemonWatcherService(cotyledon.Service):
    name = "watcher"

    def __init__(self, worker_id, registry):
        super(CNIDaemonWatcherService, self).__init__(worker_id)
        self.registry = registry
        self.host = CONF.host
        self.context = zun_context.get_admin_context(all_projects=True)
        self.neutron_api = neutron.NeutronAPI(self.context)
        self.periodic = periodics.PeriodicWorker.create(
            [], executor_factory=lambda: futures.ThreadPoolExecutor(
                max_workers=1))

    def run(self):
        self.periodic.add(self.sync_containers)
        self.periodic.add(self.poll_vif_status)
        self.periodic.start()

    @periodics.periodic(spacing=60, run_immediately=True)
    def sync_containers(self):
        LOG.debug('Start syncing capsule/container states.')
        capsules = objects.Capsule.list_by_host(self.context, self.host)
        containers = objects.Container.list_by_host(self.context, self.host)
        container_in_db = set()
        for container in (capsules + containers):
            container_in_db.add(container.uuid)
        container_in_registry = self.registry.keys()
        # process capsules/containers that are deleted
        for uuid in container_in_registry:
            if uuid not in container_in_db:
                self._on_container_deleted(uuid)

    def _on_container_deleted(self, container_uuid):
        try:
            # NOTE(ndesh): We need to lock here to avoid race condition
            #              with the deletion code for CNI DEL so that
            #              we delete the registry entry exactly once
            with lockutils.lock(container_uuid, external=True):
                if self.registry[container_uuid]['vif_unplugged']:
                    LOG.debug("Remove capsule/container %(container)s from "
                              "registry", {'container': container_uuid})
                    del self.registry[container_uuid]
                else:
                    LOG.debug("Received delete for capsule/container "
                              "%(container)s", {'container': container_uuid})
                    container_dict = self.registry[container_uuid]
                    container_dict['del_received'] = True
                    self.registry[container_uuid] = container_dict
        except KeyError:
            # This means someone else removed it. It's odd but safe to ignore.
            LOG.debug('Capsule/Container %s entry already removed from '
                      'registry while handling DELETED event. Ignoring.',
                      container_uuid)
            pass

    @periodics.periodic(spacing=1)
    def poll_vif_status(self):
        # get a copy of registry data stored in manager process
        registry_dict = self.registry.copy()
        inactive_vifs = {}
        for container_uuid in registry_dict:
            for ifname in registry_dict[container_uuid]['vifs']:
                vif_dict = registry_dict[container_uuid]['vifs'][ifname]
                if not vif_dict['active']:
                    inactive_vifs[vif_dict['id']] = ifname
        if not inactive_vifs:
            return

        LOG.debug('Checking status of vifs: %s', inactive_vifs.keys())
        # TODO(hongbin): search ports by device_owner as well
        search_opts = {'binding:host_id': self.host}
        ports = self.neutron_api.list_ports(**search_opts)['ports']
        for container_uuid in registry_dict:
            for port in ports:
                port_id = port['id']
                if port_id in inactive_vifs and utils.is_port_active(port):
                    ifname = inactive_vifs[port_id]
                    LOG.debug('sync status of port: %s', port_id)
                    self._update_vif_status(container_uuid, ifname)

    def _update_vif_status(self, container_uuid, ifname):
        with lockutils.lock(container_uuid, external=True):
            container_dict = self.registry.get(container_uuid)
            if container_dict:
                container_dict = self.registry[container_uuid]
                container_dict['vifs'][ifname]['active'] = True
                self.registry[container_uuid] = container_dict

    def terminate(self):
        if self.periodic:
            self.periodic.stop()


class CNIDaemonServiceManager(cotyledon.ServiceManager):
    def __init__(self):
        super(CNIDaemonServiceManager, self).__init__()
        # TODO(dulek): Use cotyledon.oslo_config_glue to support conf reload.
        self.manager = multiprocessing.Manager()
        registry = self.manager.dict()  # For Watcher->Server communication.
        self.add(CNIDaemonWatcherService, workers=1, args=(registry,))
        self.add(CNIDaemonServerService, workers=1, args=(registry,))
        self.register_hooks(on_terminate=self.terminate)

    def run(self):
        reaper_thread = threading.Thread(target=self._zombie_reaper,
                                         daemon=True)
        self._terminate_called = threading.Event()
        reaper_thread.start()
        super(CNIDaemonServiceManager, self).run()

    def _zombie_reaper(self):
        while True:
            try:
                res = os.waitpid(-1, os.WNOHANG)
                # don't sleep or stop if a zombie process was found
                # as there could be more
                if res != (0, 0):
                    continue
            except ChildProcessError:
                # There are no child processes yet (or they have been killed)
                pass
            except os.error:
                LOG.exception("Got OS error while reaping zombie processes")
            if self._terminate_called.isSet():
                break
            time.sleep(1)

    def terminate(self):
        self._terminate_called.set()
        self.manager.shutdown()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from zun.common import exception
from zun.common import utils
import zun.conf
from zun.tests import base
from zun.volume import driver


CONF = zun.conf.CONF


class CinderVolumeDriverTestCase(base.TestCase):

    def setUp(self):
        super(CinderVolumeDriverTestCase, self).setUp()
        self.fake_uuid = uuidutils.generate_uuid()
        self.fake_volume_id = 'fake-volume-id'
        self.fake_devpath = '/fake-path'
        self.fake_mountpoint = '/fake-mountpoint'
        self.fake_container_path = '/fake-container-path'
        self.fake_conn_info = {
            'data': {'device_path': self.fake_devpath},
        }
        self.volmap = mock.MagicMock()
        self.volmap.volume.uuid = self.fake_uuid
        self.volmap.volume_provider = 'cinder'
        self.volmap.volume_id = self.fake_volume_id
        self.volmap.container_path = self.fake_container_path
        self.volmap.connection_info = jsonutils.dumps(self.fake_conn_info)

    @mock.patch('zun.common.mount.do_mount')
    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('zun.common.mount.get_mountpoint')
    @mock.patch('zun.volume.cinder_workflow.CinderWorkflow')
    def test_attach(self, mock_cinder_workflow_cls, mock_get_mountpoint,
                    mock_ensure_tree, mock_do_mount):
        mock_cinder_workflow = mock.MagicMock()
        mock_cinder_workflow_cls.return_value = mock_cinder_workflow
        mock_cinder_workflow.attach_volume.return_value = self.fake_devpath
        mock_get_mountpoint.return_value = self.fake_mountpoint

        volume_driver = driver.Cinder()
        self.volmap.connection_info = None
        volume_driver.attach(self.context, self.volmap)

        mock_cinder_workflow.attach_volume.assert_called_once_with(self.volmap)
        mock_get_mountpoint.assert_called_once_with(self.fake_uuid)
        mock_do_mount.assert_called_once_with(
            self.fake_devpath, self.fake_mountpoint, CONF.volume.fstype)
        mock_cinder_workflow.detach_volume.assert_not_called()

    @mock.patch('zun.common.mount.do_mount')
    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('zun.common.mount.get_mountpoint')
    @mock.patch('zun.volume.cinder_workflow.CinderWorkflow')
    def test_attach_unknown_provider(self, mock_cinder_workflow_cls,
                                     mock_get_mountpoint, mock_ensure_tree,
                                     mock_do_mount):
        volume_driver = driver.Cinder()
        self.volmap.volume_provider = 'unknown'
        self.assertRaises(exception.ZunException,
                          volume_driver.attach, self.context, self.volmap)

    @mock.patch('zun.common.mount.do_mount')
    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('zun.common.mount.get_mountpoint')
    @mock.patch('zun.volume.cinder_workflow.CinderWorkflow')
    def test_attach_fail_attach(self, mock_cinder_workflow_cls,
                                mock_get_mountpoint, mock_ensure_tree,
                                mock_do_mount):
        mock_cinder_workflow = mock.MagicMock()
        mock_cinder_workflow_cls.return_value = mock_cinder_workflow
        mock_cinder_workflow.attach_volume.side_effect = \
            exception.ZunException()
        mock_get_mountpoint.return_value = self.fake_mountpoint

        volume_driver = driver.Cinder()
        self.volmap.connection_info = None
        self.assertRaises(exception.ZunException,
                          volume_driver.attach, self.context, self.volmap)

        mock_cinder_workflow.attach_volume.assert_called_once_with(self.volmap)
        mock_get_mountpoint.assert_not_called()
        mock_do_mount.assert_not_called()
        mock_cinder_workflow.detach_volume.assert_not_called()

    @mock.patch('zun.common.mount.do_mount')
    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('zun.common.mount.get_mountpoint')
    @mock.patch('zun.volume.cinder_workflow.CinderWorkflow')
    def test_attach_fail_mount(self, mock_cinder_workflow_cls,
                               mock_get_mountpoint, mock_ensure_tree,
                               mock_do_mount):
        mock_cinder_workflow = mock.MagicMock()
        mock_cinder_workflow_cls.return_value = mock_cinder_workflow
        mock_cinder_workflow.attach_volume.return_value = self.fake_devpath
        mock_get_mountpoint.return_value = self.fake_mountpoint
        mock_do_mount.side_effect = exception.ZunException()

        volume_driver = driver.Cinder()
        self.volmap.connection_info = None
        self.assertRaises(exception.ZunException,
                          volume_driver.attach, self.context, self.volmap)

        mock_cinder_workflow.attach_volume.assert_called_once_with(self.volmap)
        mock_get_mountpoint.assert_called_once_with(self.fake_uuid)
        mock_do_mount.assert_called_once_with(
            self.fake_devpath, self.fake_mountpoint, CONF.volume.fstype)
        mock_cinder_workflow.detach_volume.assert_called_once_with(
            self.context, self.volmap)

    @mock.patch('zun.common.mount.do_mount')
    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('zun.common.mount.get_mountpoint')
    @mock.patch('zun.volume.cinder_workflow.CinderWorkflow')
    def test_attach_fail_mount_and_detach(self, mock_cinder_workflow_cls,
                                          mock_get_mountpoint,
                                          mock_ensure_tree,
                                          mock_do_mount):
        class TestException1(Exception):
            pass

        class TestException2(Exception):
            pass

        mock_cinder_workflow = mock.MagicMock()
        mock_cinder_workflow_cls.return_value = mock_cinder_workflow
        mock_cinder_workflow.attach_volume.return_value = self.fake_devpath
        mock_get_mountpoint.return_value = self.fake_mountpoint
        mock_do_mount.side_effect = TestException1()
        mock_cinder_workflow.detach_volume.side_effect = TestException2()

        volume_driver = driver.Cinder()
        self.volmap.connection_info = None
        self.assertRaises(TestException1,
                          volume_driver.attach, self.context, self.volmap)

        mock_cinder_workflow.attach_volume.assert_called_once_with(self.volmap)
        mock_get_mountpoint.assert_called_once_with(self.fake_uuid)
        mock_do_mount.assert_called_once_with(
            self.fake_devpath, self.fake_mountpoint, CONF.volume.fstype)
        mock_cinder_workflow.detach_volume.assert_called_once_with(
            self.context, self.volmap)

    @mock.patch('shutil.rmtree')
    @mock.patch('zun.common.mount.do_unmount')
    @mock.patch('zun.common.mount.get_mountpoint')
    @mock.patch('zun.volume.cinder_workflow.CinderWorkflow')
    def test_detach(self, mock_cinder_workflow_cls, mock_get_mountpoint,
                    mock_do_unmount, mock_rmtree):
        mock_cinder_workflow = mock.MagicMock()
        mock_cinder_workflow_cls.return_value = mock_cinder_workflow
        mock_get_mountpoint.return_value = self.fake_mountpoint

        volume_driver = driver.Cinder()
        volume_driver.detach(self.context, self.volmap)
        mock_cinder_workflow.detach_volume.\
            assert_called_once_with(self.context, self.volmap)
        mock_get_mountpoint.assert_called_once_with(self.fake_uuid)
        mock_do_unmount.assert_called_once_with(self.fake_mountpoint)
        mock_rmtree.assert_called_once_with(self.fake_mountpoint)

    @mock.patch('zun.common.mount.get_mountpoint')
    @mock.patch('zun.volume.cinder_workflow.CinderWorkflow')
    def test_bind_mount(self, mock_cinder_workflow_cls, mock_get_mountpoint):
        mock_cinder_workflow = mock.MagicMock()
        mock_cinder_workflow_cls.return_value = mock_cinder_workflow
        mock_get_mountpoint.return_value = self.fake_mountpoint

        volume_driver = driver.Cinder()
        source, destination = volume_driver.bind_mount(
            self.context, self.volmap)

        self.assertEqual(self.fake_mountpoint, source)
        self.assertEqual(self.fake_container_path, destination)
        mock_get_mountpoint.assert_called_once_with(self.fake_uuid)

    @mock.patch('zun.common.mount.get_mountpoint')
    @mock.patch('zun.common.mount.Mounter.read_mounts')
    @mock.patch('zun.volume.cinder_workflow.CinderWorkflow')
    def test_delete(self, mock_cinder_workflow_cls, mock_read_mounts,
                    mock_get_mountpoint):
        mock_cinder_workflow = mock.MagicMock()
        mock_cinder_workflow_cls.return_value = mock_cinder_workflow
        mock_cinder_workflow.delete_volume.return_value = self.fake_volume_id
        mock_get_mountpoint.return_value = self.fake_mountpoint

        volume_driver = driver.Cinder()
        volume_driver.delete(self.context, self.volmap)

        mock_cinder_workflow.delete_volume.assert_called_once_with(self.volmap)


class LocalVolumeDriverTestCase(base.TestCase):

    def setUp(self):
        super(LocalVolumeDriverTestCase, self).setUp()
        self.fake_uuid = uuidutils.generate_uuid()
        self.fake_mountpoint = '/fake-mountpoint'
        self.fake_container_path = '/fake-container-path'
        self.fake_contents = 'fake-contents'
        self.volmap = mock.MagicMock()
        self.volmap.volume.uuid = self.fake_uuid
        self.volmap.volume_provider = 'local'
        self.volmap.container_path = self.fake_container_path
        self.volmap.contents = self.fake_contents

    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('zun.common.mount.get_mountpoint')
    def test_attach(self, mock_get_mountpoint, mock_ensure_tree):
        mock_get_mountpoint.return_value = self.fake_mountpoint
        volume_driver = driver.Local()

        with mock.patch('zun.volume.driver.open', mock.mock_open()
                        ) as mock_open:
            volume_driver.attach(self.context, self.volmap)

        expected_file_path = self.fake_mountpoint + '/' + self.fake_uuid
        mock_open.assert_called_once_with(expected_file_path, 'wb')
        mock_open().write.assert_called_once_with(
            utils.decode_file_data(self.fake_contents))
        mock_get_mountpoint.assert_called_once_with(self.fake_uuid)

    @mock.patch('shutil.rmtree')
    @mock.patch('zun.common.mount.get_mountpoint')
    def test_detach(self, mock_get_mountpoint, mock_rmtree):
        mock_get_mountpoint.return_value = self.fake_mountpoint
        volume_driver = driver.Local()
        volume_driver.detach(self.context, self.volmap)

        mock_get_mountpoint.assert_called_once_with(self.fake_uuid)
        mock_rmtree.assert_called_once_with(self.fake_mountpoint)

    @mock.patch('zun.common.mount.get_mountpoint')
    def test_bind_mount(self, mock_get_mountpoint):
        mock_get_mountpoint.return_value = self.fake_mountpoint
        volume_driver = driver.Local()
        source, destination = volume_driver.bind_mount(
            self.context, self.volmap)

        expected_file_path = self.fake_mountpoint + '/' + self.fake_uuid
        self.assertEqual(expected_file_path, source)
        self.assertEqual(self.fake_container_path, destination)
        mock_get_mountpoint.assert_called_once_with(self.fake_uuid)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Create quota & quota class tables

Revision ID: 2b045cb595db
Revises: 238f94009eab
Create Date: 2018-04-09 13:33:52.522262

"""

# revision identifiers, used by Alembic.
revision = '2b045cb595db'
down_revision = '238f94009eab'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'quotas',
        sa.Column('created_at', sa.DateTime()),
        sa.Column('updated_at', sa.DateTime()),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('project_id', sa.String(length=255), index=True),
        sa.Column('resource', sa.String(length=255), nullable=False),
        sa.Column('hard_limit', sa.Integer()),
        sa.PrimaryKeyConstraint('id'),
        mysql_charset='utf8',
        mysql_engine='InnoDB'
    )

    op.create_table(
        'quota_classes',
        sa.Column('created_at', sa.DateTime()),
        sa.Column('updated_at', sa.DateTime()),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('class_name', sa.String(length=255)),
        sa.Column('resource', sa.String(length=255)),
        sa.Column('hard_limit', sa.Integer()),
        sa.Index('quota_classes_class_name_idx', 'class_name'),
        sa.PrimaryKeyConstraint('id'),
        mysql_charset='utf8',
        mysql_engine='InnoDB'
    )
    # ### end Alembic commands ###



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import signal
import sys

import os_vif
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from zun.cni import api as cni_api
from zun.cni import utils
from zun.common import config
from zun.common import consts


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
_CNI_TIMEOUT = 180


def main():
    d = jsonutils.load(sys.stdin.buffer)
    cni_conf = utils.CNIConfig(d)
    args = (['--config-file', cni_conf.zun_conf] if 'zun_conf' in d
            else [])

    try:
        if cni_conf.debug:
            args.append('-d')
    except AttributeError:
        pass
    config.init(args)
    if os.environ.get('CNI_COMMAND') == 'VERSION':
        CONF.set_default('use_stderr', True)

    # Initialize o.vo registry.
    os_vif.initialize()

    runner = cni_api.CNIDaemonizedRunner()

    def _timeout(signum, frame):
        runner._write_dict(sys.stdout, {
            'msg': 'timeout',
            'code': consts.CNI_TIMEOUT_CODE,
        })
        LOG.debug('timed out')
        sys.exit(1)

    signal.signal(signal.SIGALRM, _timeout)
    signal.alarm(_CNI_TIMEOUT)
    status = runner.run(os.environ, cni_conf, sys.stdout)
    LOG.debug("Exiting with status %s", status)
    if status:
        sys.exit(status)



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock
from unittest.mock import patch

from oslo_utils import uuidutils
from webtest.app import AppError

from zun import objects
from zun.tests.unit.api import base as api_base
from zun.tests.unit.db import utils
from zun.tests.unit.objects import utils as obj_utils


class TestRegistryController(api_base.FunctionalTest):

    def test_create_registry(self):
        params = ('{"registry": {"name": "MyRegistry", "domain": "test.io",'
                  '"username": "fake-user", "password": "fake-pass"}}')
        response = self.post('/v1/registries/',
                             params=params,
                             content_type='application/json')

        self.assertEqual(201, response.status_int)
        self.assertEqual(1, len(response.json))
        r = response.json['registry']
        self.assertIsNotNone(r.get('uuid'))
        self.assertIsNotNone(r.get('user_id'))
        self.assertIsNotNone(r.get('project_id'))
        self.assertEqual('MyRegistry', r.get('name'))
        self.assertEqual('test.io', r.get('domain'))
        self.assertEqual('fake-user', r.get('username'))
        self.assertEqual('***', r.get('password'))

    def test_create_registry_domain_not_specified(self):
        params = ('{"registry": {"name": "MyRegistry",'
                  '"username": "fake-user", "password": "fake-pass"}}')
        with self.assertRaisesRegex(AppError,
                                    "is a required property"):
            self.post('/v1/registries/',
                      params=params,
                      content_type='application/json')

    def test_create_registry_with_minimum_params(self):
        params = ('{"registry": {"domain": "test.io"}}')
        response = self.post('/v1/registries/',
                             params=params,
                             content_type='application/json')

        self.assertEqual(201, response.status_int)
        self.assertEqual(1, len(response.json))
        r = response.json['registry']
        self.assertIsNotNone(r.get('uuid'))
        self.assertIsNotNone(r.get('user_id'))
        self.assertIsNotNone(r.get('project_id'))
        self.assertIsNone(r.get('name'))
        self.assertEqual('test.io', r.get('domain'))
        self.assertIsNone(r.get('username'))
        self.assertEqual('***', r.get('password'))

    def test_create_registry_invalid_long_name(self):
        # Long name
        params = ('{"registry": {"name": "' + 'i' * 256 + '",'
                  '"domain": "test.io","username": "fake-user",'
                  '"password": "fake-pass"}}')
        self.assertRaises(AppError, self.post, '/v1/registries/',
                          params=params, content_type='application/json')

    def test_get_all_registries(self):
        params = ('{"registry": {"name": "MyRegistry", "domain": "test.io",'
                  '"username": "fake-user", "password": "fake-pass"}}')
        response = self.post('/v1/registries/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(201, response.status_int)
        response = self.get('/v1/registries/')
        self.assertEqual(200, response.status_int)
        self.assertEqual(2, len(response.json))
        r = response.json['registries'][0]
        self.assertIsNotNone(r.get('uuid'))
        self.assertIsNotNone(r.get('user_id'))
        self.assertIsNotNone(r.get('project_id'))
        self.assertEqual('MyRegistry', r.get('name'))
        self.assertEqual('test.io', r.get('domain'))
        self.assertEqual('fake-user', r.get('username'))
        self.assertEqual('***', r.get('password'))

    @patch('zun.common.policy.enforce')
    @patch('zun.objects.Registry.list')
    def test_get_all_registries_all_projects(self, mock_registry_list,
                                             mock_policy):
        mock_policy.return_value = True
        test_registry = utils.get_test_registry()
        registries = [objects.Registry(self.context, **test_registry)]
        mock_registry_list.return_value = registries

        response = self.get('/v1/registries/?all_projects=1')

        mock_registry_list.assert_called_once_with(mock.ANY,
                                                   1000, None, 'id', 'asc',
                                                   filters={})
        context = mock_registry_list.call_args[0][0]
        self.assertIs(True, context.all_projects)
        self.assertEqual(200, response.status_int)
        actual_registries = response.json['registries']
        self.assertEqual(1, len(actual_registries))
        self.assertEqual(test_registry['uuid'],
                         actual_registries[0].get('uuid'))

    @patch('zun.objects.Registry.list')
    def test_get_all_registries_with_pagination_marker(self,
                                                       mock_registry_list):
        registry_list = []
        for id_ in range(4):
            test_registry = utils.create_test_registry(
                id=id_, uuid=uuidutils.generate_uuid(),
                name='registry' + str(id_), context=self.context)
            registry_list.append(objects.Registry(self.context,
                                                  **test_registry))
        mock_registry_list.return_value = registry_list[-1:]
        response = self.get('/v1/registries/?limit=3&marker=%s'
                            % registry_list[2].uuid)

        self.assertEqual(200, response.status_int)
        actual_registries = response.json['registries']
        self.assertEqual(1, len(actual_registries))
        self.assertEqual(registry_list[-1].uuid,
                         actual_registries[0].get('uuid'))

    @patch('zun.objects.Registry.list')
    def test_get_all_registries_with_filter(self, mock_registry_list):
        test_registry = utils.get_test_registry()
        registries = [objects.Registry(self.context, **test_registry)]
        mock_registry_list.return_value = registries

        response = self.get('/v1/registries/?name=fake-name')

        mock_registry_list.assert_called_once_with(
            mock.ANY, 1000, None, 'id', 'asc', filters={'name': 'fake-name'})
        self.assertEqual(200, response.status_int)
        actual_registries = response.json['registries']
        self.assertEqual(1, len(actual_registries))
        self.assertEqual(test_registry['uuid'],
                         actual_registries[0].get('uuid'))

    @patch('zun.objects.Registry.list')
    def test_get_all_registries_with_unknown_parameter(
            self, mock_registry_list):
        test_registry = utils.get_test_registry()
        registries = [objects.Registry(self.context, **test_registry)]
        mock_registry_list.return_value = registries

        response = self.get('/v1/registries/?unknown=fake-name',
                            expect_errors=True)

        mock_registry_list.assert_not_called()
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual("Unknown parameters: unknown",
                         response.json['errors'][0]['detail'])

    def test_get_one(self):
        params = ('{"registry": {"name": "MyRegistry", "domain": "test.io",'
                  '"username": "fake-user", "password": "fake-pass"}}')
        response = self.post('/v1/registries/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(201, response.status_int)
        # get by uuid
        registry_uuid = response.json['registry']['uuid']
        response = self.get('/v1/registries/%s/' % registry_uuid)
        self.assertEqual(200, response.status_int)
        self.assertEqual(1, len(response.json))
        r = response.json['registry']
        self.assertIsNotNone(r.get('uuid'))
        self.assertIsNotNone(r.get('user_id'))
        self.assertIsNotNone(r.get('project_id'))
        self.assertEqual('MyRegistry', r.get('name'))
        self.assertEqual('test.io', r.get('domain'))
        self.assertEqual('fake-user', r.get('username'))
        self.assertEqual('***', r.get('password'))
        # get by name
        registry_name = response.json['registry']['name']
        response = self.get('/v1/registries/%s/' % registry_name)
        self.assertEqual(200, response.status_int)
        self.assertEqual(1, len(response.json))
        r = response.json['registry']
        self.assertIsNotNone(r.get('uuid'))
        self.assertIsNotNone(r.get('user_id'))
        self.assertIsNotNone(r.get('project_id'))
        self.assertEqual('MyRegistry', r.get('name'))
        self.assertEqual('test.io', r.get('domain'))
        self.assertEqual('fake-user', r.get('username'))
        self.assertEqual('***', r.get('password'))

    def test_get_one_not_found(self):
        response = self.get('/v1/registries/%s/' % 'not-exist',
                            expect_errors=True)

        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual("Registry not-exist could not be found.",
                         response.json['errors'][0]['detail'])

    def test_patch_by_uuid(self):
        params = ('{"registry": {"name": "MyRegistry", "domain": "test.io",'
                  '"username": "fake-user", "password": "fake-pass"}}')
        response = self.post('/v1/registries/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(201, response.status_int)
        registry_uuid = response.json['registry']['uuid']
        params = {'registry': {'name': 'new-name', 'domain': 'new-domain',
                  'username': 'new-username', 'password': 'new-pass'}}
        response = self.patch_json(
            '/registries/%s/' % registry_uuid,
            params=params)
        self.assertEqual(200, response.status_int)
        self.assertEqual(1, len(response.json))
        r = response.json['registry']
        self.assertIsNotNone(r.get('uuid'))
        self.assertIsNotNone(r.get('user_id'))
        self.assertIsNotNone(r.get('project_id'))
        self.assertEqual('new-name', r.get('name'))
        self.assertEqual('new-domain', r.get('domain'))
        self.assertEqual('new-username', r.get('username'))
        self.assertEqual('***', r.get('password'))

    def test_delete_registry_by_uuid(self):
        params = ('{"registry": {"name": "MyRegistry", "domain": "test.io",'
                  '"username": "fake-user", "password": "fake-pass"}}')
        response = self.post('/v1/registries/',
                             params=params,
                             content_type='application/json')
        self.assertEqual(201, response.status_int)
        registry_uuid = response.json['registry']['uuid']
        response = self.delete('/v1/registries/%s/' % registry_uuid)
        self.assertEqual(204, response.status_int)
        response = self.get('/v1/registries/%s/' % registry_uuid,
                            expect_errors=True)
        self.assertEqual(404, response.status_int)


class TestRegistryEnforcement(api_base.FunctionalTest):

    def _common_policy_check(self, rule, func, *arg, **kwarg):
        rules = dict({rule: 'project_id:non_fake'},
                     **kwarg.pop('bypass_rules', {}))
        self.policy.set_rules(rules)
        response = func(*arg, **kwarg)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_disallow_get_all(self):
        self._common_policy_check(
            'registry:get_all', self.get, '/v1/registries/',
            expect_errors=True)

    def test_policy_disallow_get_all_all_projects(self):
        self._common_policy_check(
            'registry:get_all_all_projects',
            self.get, '/v1/registries/?all_projects=1',
            expect_errors=True,
            bypass_rules={'registry:get_all': 'project_id:fake_project'})

    def test_policy_disallow_get_one(self):
        registry = obj_utils.create_test_registry(self.context)
        self._common_policy_check(
            'registry:get_one', self.get,
            '/v1/registries/%s/' % registry.uuid,
            expect_errors=True)

    def test_policy_disallow_update(self):
        registry = obj_utils.create_test_registry(self.context)
        params = {'registry': {'name': 'newname'}}
        self._common_policy_check(
            'registry:update', self.patch_json,
            '/registries/%s/' % registry.uuid, params,
            expect_errors=True)

    def test_policy_disallow_create(self):
        params = ('{"registry": {"domain": "test.io"}}')
        self._common_policy_check(
            'registry:create', self.post, '/v1/registries/',
            params=params,
            content_type='application/json',
            expect_errors=True)

    def test_policy_disallow_delete(self):
        registry = obj_utils.create_test_registry(self.context)
        self._common_policy_check(
            'registry:delete', self.delete,
            '/v1/registries/%s/' % registry.uuid,
            expect_errors=True)

    def _owner_check(self, rule, func, *args, **kwargs):
        self.policy.set_rules({rule: "user_id:%(user_id)s"})
        response = func(*args, **kwargs)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_only_owner_get_one(self):
        registry = obj_utils.create_test_registry(self.context,
                                                  user_id='another')
        self._owner_check("registry:get_one", self.get_json,
                          '/registries/%s/' % registry.uuid,
                          expect_errors=True)

    def test_policy_only_owner_update(self):
        registry = obj_utils.create_test_registry(self.context,
                                                  user_id='another')
        self._owner_check(
            "registry:update", self.patch_json,
            '/registries/%s/' % registry.uuid,
            {'registry': {'name': 'newname'}},
            expect_errors=True)

    def test_policy_only_owner_delete(self):
        registry = obj_utils.create_test_registry(self.context,
                                                  user_id='another')
        self._owner_check(
            "registry:delete", self.delete,
            '/registries/%s/' % registry.uuid,
            expect_errors=True)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from stevedore import driver as stevedore_driver

import zun.conf


CONF = zun.conf.CONF


def driver(*args, **kwargs):
    driver_name = CONF.network.driver
    network_driver = stevedore_driver.DriverManager(
        "zun.network.driver",
        driver_name,
        invoke_on_load=True).driver

    network_driver.init(*args, **kwargs)
    return network_driver


class Network(object, metaclass=abc.ABCMeta):
    """The base class that all Network classes should inherit from."""

    def init(self, context, *args, **kwargs):
        raise NotImplementedError()

    def get_or_create_network(self, *args, **kwargs):
        raise NotImplementedError()

    def create_network(self, *args, **kwargs):
        raise NotImplementedError()

    def remove_network(self, network_name, **kwargs):
        raise NotImplementedError()

    def process_networking_config(self, *args, **kwargs):
        raise NotImplementedError()

    def connect_container_to_network(self, container, network_name, **kwargs):
        raise NotImplementedError()

    def disconnect_container_from_network(self, *args, **kwargs):
        raise NotImplementedError()

    def add_security_groups_to_ports(self, *args, **kwargs):
        raise NotImplementedError()

    def remove_security_groups_from_ports(self, container, security_group_ids,
                                          **kwargs):
        raise NotImplementedError()

    def on_container_started(self, container):
        raise NotImplementedError()

    def on_container_stopped(self, container):
        raise NotImplementedError()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

CONTAINER_STATUSES = (
    ERROR, RUNNING, STOPPED, PAUSED, UNKNOWN, CREATING, CREATED,
    DELETED, DELETING, REBUILDING, DEAD, RESTARTING
) = (
    'Error', 'Running', 'Stopped', 'Paused', 'Unknown', 'Creating', 'Created',
    'Deleted', 'Deleting', 'Rebuilding', 'Dead', 'Restarting'
)

CAPSULE_STATUSES = (
    PENDING, RUNNING, SUCCEEDED, FAILED, UNKNOWN
) = (
    'Pending', 'Running', 'Succeeded', 'Failed', 'Unknown'
)

TASK_STATES = (
    IMAGE_PULLING, CONTAINER_CREATING,
    CONTAINER_STARTING, CONTAINER_DELETING,
    CONTAINER_STOPPING, CONTAINER_REBOOTING, CONTAINER_PAUSING,
    CONTAINER_UNPAUSING, CONTAINER_KILLING, SG_ADDING,
    SG_REMOVING, NETWORK_ATTACHING, NETWORK_DETACHING,
    CONTAINER_REBUILDING,
) = (
    'image_pulling', 'container_creating',
    'container_starting', 'container_deleting',
    'container_stopping', 'container_rebooting', 'container_pausing',
    'container_unpausing', 'container_killing', 'sg_adding',
    'sg_removing', 'network_attaching', 'network_detaching',
    'container_rebuilding',
)

RESOURCE_CLASSES = (
    VCPU, MEMORY_MB, DISK_GB, PCI_DEVICE, SRIOV_NET_VF,
    NUMA_SOCKET, NUMA_CORE, NUMA_THREAD, NUMA_MEMORY_MB,
    IPV4_ADDRESS
) = (
    'VCPU', 'MEMORY_MB', 'DISK_GB', 'PCI_DEVICE', 'SRIOV_NET_VF',
    'NUMA_SOCKET', 'NUMA_CORE', 'NUMA_THREAD', 'NUMA_MEMORY_MB',
    'IPV4_ADDRESS'
)

ALLOCATED = 'allocated'

# The name of Docker container is of the form NAME_PREFIX-<uuid>
NAME_PREFIX = 'zun-'

# Storage drivers that support disk quota feature
SUPPORTED_STORAGE_DRIVERS = \
    ['devicemapper', 'overlay2', 'windowfilter', 'zfs', 'btrfs']

DEFAULT_QUOTA_CLASS_NAME = 'default'

TYPE_ANY = -1
TYPE_CONTAINER = 0
TYPE_CAPSULE = 1
TYPE_CAPSULE_CONTAINER = 2
TYPE_CAPSULE_INIT_CONTAINER = 3

CUSTOM_TRAITS = (
    ZUN_COMPUTE_STATUS_DISABLED,
) = (
    'CUSTOM_ZUN_COMPUTE_STATUS_DISABLED',
)

# neutron related constants
BINDING_PROFILE = 'binding:profile'
BINDING_HOST_ID = 'binding:host_id'
DEVICE_OWNER_ZUN = 'compute:zun'

# CNI constants
CNI_EXCEPTION_CODE = 100
CNI_TIMEOUT_CODE = 200
DEFAULT_IFNAME = 'eth0'
CNI_METADATA_VIF = 'vif'
CNI_METADATA_PID = 'pid'
USERSPACE_DRIVERS = ['vfio-pci', 'uio', 'uio_pci_generic', 'igb_uio']



# Copyright 2017 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import builtins
from unittest import mock
from unittest.mock import mock_open

from oslo_serialization import jsonutils

from zun.common import exception
from zun.container.os_capability.linux import os_capability_linux
from zun.tests import base

LSCPU_ON = """# The following is the parsable format, which can be fed to other
# programs. Each different item in every column has an unique ID
# starting from zero.
# Socket,CPU,Online
0,0,Y
0,8,Y
1,16,Y
1,24,Y
2,32,Y"""

LSCPU_NO_ONLINE = """# The following is the parsable format, which can be fed to
# programs. Each different item in every column has an unique ID
# starting from zero.
# Socket,CPU
0,0
0,1
1,2
1,3"""


class TestOSCapability(base.BaseTestCase):
    @mock.patch('zun.common.utils.execute')
    def test_get_cpu_numa_info_with_online(self, mock_output):
        mock_output.return_value = LSCPU_ON
        output = os_capability_linux.LinuxHost().get_cpu_numa_info()
        expected_output = {'0': [0, 8], '1': [16, 24], '2': [32]}
        self.assertEqual(expected_output, output)

    @mock.patch('zun.common.utils.execute')
    def test_get_cpu_numa_info_exception(self, mock_output):
        mock_output.side_effect = exception.CommandError()
        self.assertRaises(exception.CommandError,
                          os_capability_linux.LinuxHost().get_cpu_numa_info)

    @mock.patch('zun.common.utils.execute')
    def test_get_cpu_numa_info_without_online(self, mock_output):
        mock_output.side_effect = [exception.CommandError(),
                                   LSCPU_NO_ONLINE]
        expected_output = {'0': [0, 1], '1': [2, 3]}
        output = os_capability_linux.LinuxHost().get_cpu_numa_info()
        self.assertEqual(expected_output, output)

    def test_get_host_mem(self):
        data = ('MemTotal:        3882464 kB\nMemFree:         3514608 kB\n'
                'MemAvailable:    3556372 kB\n')
        m_open = mock_open(read_data=data)
        with mock.patch.object(builtins, "open", m_open,
                               create=True):
            output = os_capability_linux.LinuxHost().get_host_mem()
            used = (3882464 - 3556372)
            self.assertEqual((3882464, 3514608, 3556372, used), output)

    @mock.patch('zun.pci.utils.get_ifname_by_pci_address')
    @mock.patch('zun.pci.utils.get_net_name_by_vf_pci_address')
    @mock.patch('zun.common.utils.execute')
    def test_get_pci_resource(self, mock_output, mock_netname,
                              mock_ifname):
        mock_netname.return_value = 'net_enp2s0f3_ec_38_8f_79_11_2b'
        mock_ifname.return_value = 'enp2s0f3'
        value1 = '''0000:02:10.7 "Ethernet controller...." ""'''
        value2 = '02:10.7 0200: 8086:1520 (rev 01)'
        value3 = '''Slot:   02:10.7
        Class:  Ethernet controller
        Vendor: Intel Corporation
        Device: I350 Ethernet Controller Virtual Function
        Rev:    01
        NUMANode:   0'''
        value4 = 'class physfn'
        value5 = '''DRIVER=igbvf
        PCI_CLASS=20000
        PCI_ID=8086:1520
        PCI_SUBSYS_ID=FFFF:0000
        PCI_SLOT_NAME=0000:02:10.7
        MODALIAS=pci:v00008086d00001520sv0000FFFFsd00000000bc02sc00i00'''
        value6 = '''Features for enp2s0f3:
rx-checksumming: on
tx-checksumming: on
scatter-gather: on
tcp-segmentation-offload: on
generic-receive-offload: on
large-receive-offload: off [fixed]
rx-vlan-offload: on
tx-vlan-offload: on
ntuple-filters: off [fixed]
receive-hashing: on
highdma: on [fixed]
rx-vlan-filter: on [fixed]
vlan-challenged: off [fixed]
tx-lockless: off [fixed]
netns-local: off [fixed]
tx-gso-robust: off [fixed]
tx-fcoe-segmentation: off [fixed]
tx-gre-segmentation: off [fixed]
tx-ipip-segmentation: off [fixed]
tx-sit-segmentation: off [fixed]
tx-udp_tnl-segmentation: off [fixed]
tx-mpls-segmentation: off [fixed]
rx-fcs: off [fixed]
tx-vlan-stag-hw-insert: off [fixed]
rx-vlan-stag-hw-parse: off [fixed]
rx-vlan-stag-filter: off [fixed]'''
        values = [(value1, 0),
                  (value2, 0),
                  (value3, 0),
                  (value4, 0),
                  (value5, 0),
                  (value6, 0)]
        mock_output.side_effect = values
        expected = {"dev_id": "pci_0000_02_10_7",
                    "address": "0000:02:10.7",
                    "vendor_id": "8086",
                    "product_id": "1520",
                    "numa_node": 0,
                    "label": "label_8086_1520",
                    "dev_type": "VF",
                    "parent_addr": "0000:02:10.7"}
        output = os_capability_linux.LinuxHost().get_pci_resources()

        pci_infos = jsonutils.loads(output)
        for pci_info in pci_infos:
            self.assertEqual(expected['dev_id'], str(pci_info['dev_id']))
            self.assertEqual(expected['address'], str(pci_info['address']))
            self.assertEqual(expected['product_id'],
                             str(pci_info['product_id']))
            self.assertEqual(expected['vendor_id'], str(pci_info['vendor_id']))
            self.assertEqual(expected['numa_node'], pci_info['numa_node'])
            self.assertEqual(expected['label'], str(pci_info['label']))
            self.assertEqual(expected['dev_type'], str(pci_info['dev_type']))
            self.assertEqual(expected['parent_addr'],
                             str(pci_info['parent_addr']))



# Copyright (c) 2017 Intel, Inc.
# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import glob
import os
import re

from oslo_log import log as logging


from zun.common import exception

LOG = logging.getLogger(__name__)

PCI_VENDOR_PATTERN = "^(hex{4})$".replace("hex", r"[\da-fA-F]")
_PCI_ADDRESS_PATTERN = ("^(hex{4}):(hex{2}):(hex{2}).(oct{1})$".
                        replace("hex", r"[\da-fA-F]").
                        replace("oct", r"[0-7]"))
_PCI_ADDRESS_REGEX = re.compile(_PCI_ADDRESS_PATTERN)

_SRIOV_TOTALVFS = "sriov_totalvfs"


def pci_device_prop_match(pci_dev, specs):
    """Check if the pci_dev meet spec requirement

    Specs is a list of PCI device property requirements.
    An example of device requirement that the PCI should be either:
    a) Device with vendor_id as 0x8086 and product_id as 0x8259, or
    b) Device with vendor_id as 0x10de and product_id as 0x10d8:

    [{"vendor_id":"8086", "product_id":"8259"},
     {"vendor_id":"10de", "product_id":"10d8",
      "capabilities_network": ["rx", "tx", "tso", "gso"]}]

    """
    def _matching_devices(spec):
        for k, v in spec.items():
            pci_dev_v = pci_dev.get(k)
            if isinstance(v, list) and isinstance(pci_dev_v, list):
                if not all(x in pci_dev.get(k) for x in v):
                    return False
            elif pci_dev_v != v:
                return False
        return True

    return any(_matching_devices(spec) for spec in specs)


def parse_address(address):
    """Returns (domain, bus, slot, function) from PCI address

    Which is stored in PciDevice DB table.
    """
    m = _PCI_ADDRESS_REGEX.match(address)
    if not m:
        raise exception.PciDeviceWrongAddressFormat(address=address)
    return m.groups()


def get_pci_address_fields(pci_addr):
    dbs, sep, func = pci_addr.partition('.')
    domain, bus, slot = dbs.split(':')
    return (domain, bus, slot, func)


def get_pci_address(domain, bus, slot, func):
    return '%s:%s:%s.%s' % (domain, bus, slot, func)


def get_function_by_ifname(ifname):
    """Get the function by the interface name

    Given the device name, returns the PCI address of a device
    and returns True if the address is in a physical function.
    """
    dev_path = "/sys/class/net/%s/device" % ifname
    sriov_totalvfs = 0
    if os.path.isdir(dev_path):
        try:
            # sriov_totalvfs contains the maximum possible VFs for this PF
            with open(os.path.join(dev_path, _SRIOV_TOTALVFS)) as fd:
                sriov_totalvfs = int(fd.read())
                return (os.readlink(dev_path).strip("./"),
                        sriov_totalvfs > 0)
        except (IOError, ValueError):
            return os.readlink(dev_path).strip("./"), False
    return None, False


def is_physical_function(domain, bus, slot, function):
    dev_path = "/sys/bus/pci/devices/%(d)s:%(b)s:%(s)s.%(f)s/" % {
        "d": domain, "b": bus, "s": slot, "f": function}
    if os.path.isdir(dev_path):
        sriov_totalvfs = 0
        try:
            with open(dev_path + _SRIOV_TOTALVFS) as fd:
                sriov_totalvfs = int(fd.read())
                return sriov_totalvfs > 0
        except (IOError, ValueError):
            pass
    return False


def _get_sysfs_netdev_path(pci_addr, pf_interface):
    """Get the sysfs path based on the PCI address of the device.

    Assumes a networking device - will not check for the existence of the path.
    """
    if pf_interface:
        return "/sys/bus/pci/devices/%s/physfn/net" % (pci_addr)
    return "/sys/bus/pci/devices/%s/net" % (pci_addr)


def get_ifname_by_pci_address(pci_addr, pf_interface=False):
    """Get the interface name based on a VF's pci address

    The returned interface name is either the parent PF's or that of the VF
    itself based on the argument of pf_interface.
    """
    dev_path = _get_sysfs_netdev_path(pci_addr, pf_interface)
    try:
        dev_info = os.listdir(dev_path)
        return dev_info.pop()
    except Exception:
        raise exception.PciDeviceNotFoundById(id=pci_addr)


def get_mac_by_pci_address(pci_addr, pf_interface=False):
    """Get the MAC address of the nic based on it's PCI address

    Raises PciDeviceNotFoundById in case the pci device is not a NIC
    """
    dev_path = _get_sysfs_netdev_path(pci_addr, pf_interface)
    if_name = get_ifname_by_pci_address(pci_addr, pf_interface)
    addr_file = os.path.join(dev_path, if_name, 'address')

    try:
        with open(addr_file) as f:
            mac = next(f).strip()
            return mac
    except (IOError, StopIteration) as e:
        LOG.warning("Could not find the expected sysfs file for "
                    "determining the MAC address of the PCI device "
                    "%(addr)s. May not be a NIC. Error: %(e)s",
                    {'addr': pci_addr, 'e': e})
        raise exception.PciDeviceNotFoundById(id=pci_addr)


def get_vf_num_by_pci_address(pci_addr):
    """Get the VF number based on a VF's pci address

    A VF is associated with an VF number, which ip link command uses to
    configure it. This number can be obtained from the PCI device filesystem.
    """
    VIRTFN_RE = re.compile(r"virtfn(\d+)")
    virtfns_path = "/sys/bus/pci/devices/%s/physfn/virtfn*" % (pci_addr)
    vf_num = None
    try:
        for vf_path in glob.iglob(virtfns_path):
            if re.search(pci_addr, os.readlink(vf_path)):
                t = VIRTFN_RE.search(vf_path)
                vf_num = t.group(1)
                break
    except Exception:
        pass
    if vf_num is None:
        raise exception.PciDeviceNotFoundById(id=pci_addr)
    return vf_num


def get_net_name_by_vf_pci_address(vfaddress):
    """Given the VF PCI address, returns the net device name and ifname."""
    try:
        mac = get_mac_by_pci_address(vfaddress).split(':')
        ifname = get_ifname_by_pci_address(vfaddress)
        return ("net_%(ifname)s_%(mac)s" %
                {'ifname': ifname, 'mac': '_'.join(mac)}), ifname
    except Exception:
        LOG.warning("No net device was found for VF %(vfaddress)s",
                    {'vfaddress': vfaddress})
        return



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import math
import sys
import time

from neutronclient.common import exceptions
from oslo_log import log as logging
from oslo_utils import excutils

from zun.common import consts
from zun.common import exception
from zun.common.i18n import _
from zun.common import utils
import zun.conf
from zun.network import network
from zun.network import neutron
from zun import objects

CONF = zun.conf.CONF

LOG = logging.getLogger(__name__)

DEVICE_OWNER = 'compute:kuryr'


class KuryrNetwork(network.Network):
    def init(self, context, docker_api):
        self.docker = docker_api
        self.neutron_api = neutron.NeutronAPI(context)
        self.context = context

    def get_or_create_network(self, context, neutron_net_id):
        docker_net_name = neutron_net_id
        docker_networks = self.docker.networks(names=[docker_net_name])
        if not docker_networks:
            self.create_network(neutron_net_id)

    def create_network(self, neutron_net_id):
        """Create a docker network with Kuryr driver.

        The docker network to be created will be based on the specified
        neutron net. It is assumed that the neutron net will have one
        or two subnets. If there are two subnets, it must be a ipv4
        subnet and a ipv6 subnet and containers created from this network
        will have both ipv4 and ipv6 addresses.

        What this method does is finding the subnets under the specified
        neutron net, retrieving the cidr, gateway of each
        subnet, and compile the list of parameters for docker.create_network.
        """
        name = neutron_net_id
        # find a v4 and/or v6 subnet of the network
        shared = \
            self.neutron_api.get_neutron_network(neutron_net_id)[
                'shared']
        subnets = self.neutron_api.list_subnets(network_id=neutron_net_id)
        subnets = subnets.get('subnets', [])
        v4_subnet = self._get_subnet(subnets, ip_version=4)
        v6_subnet = self._get_subnet(subnets, ip_version=6)
        if not v4_subnet and not v6_subnet:
            raise exception.ZunException(_(
                "The Neutron network %s has no subnet") % neutron_net_id)

        # IPAM driver specific options
        ipam_options = {
            "Driver": CONF.network.driver_name,
            "Options": {
                'neutron.net.shared': str(shared)
            },
            "Config": []
        }

        # Driver specific options
        options = {
            'neutron.net.uuid': neutron_net_id,
            'neutron.net.shared': str(shared)
        }

        if v4_subnet:
            ipam_options['Options']['neutron.subnet.uuid'] = \
                v4_subnet.get('id')
            ipam_options["Config"].append({
                "Subnet": v4_subnet['cidr'],
                "Gateway": v4_subnet['gateway_ip']
            })

            options['neutron.subnet.uuid'] = v4_subnet.get('id')
        if v6_subnet:
            ipam_options['Options']['neutron.subnet.v6.uuid'] = \
                v6_subnet.get('id')
            ipam_options["Config"].append({
                "Subnet": v6_subnet['cidr'],
                "Gateway": v6_subnet['gateway_ip']
            })

            options['neutron.subnet.v6.uuid'] = v6_subnet.get('id')

        network_dict = {}
        network_dict['project_id'] = self.context.project_id
        network_dict['user_id'] = self.context.user_id
        network_dict['name'] = name
        network_dict['neutron_net_id'] = neutron_net_id
        network = objects.ZunNetwork(self.context, **network_dict)

        for attempt in (1, 2, 3):
            LOG.debug("Attempt (%s) to create network: %s", attempt, network)
            created_network = self._create_network_attempt(
                network, options, ipam_options)
            if created_network:
                return created_network
            backoff = int(math.pow(2, attempt) - 1)
            time.sleep(backoff)

        raise exception.ZunException(_(
            "Cannot create docker network after several attempts %s"))

    def _create_network_attempt(self, network, options, ipam_options):
        # The DB model has unique constraint on 'neutron_net_id' field
        # which will guarantee only one request can create the network in here
        # (and call docker.create_network later) if there are concurrent
        # requests on creating networks for the same neutron net.
        try:
            network.create(self.context)
        except exception.NetworkAlreadyExists as e:
            if e.field != 'neutron_net_id':
                raise

            networks = objects.ZunNetwork.list(
                self.context,
                filters={'neutron_net_id': network.neutron_net_id})
            LOG.debug("network objects with 'neutron_net_id' as '%(net_id)s': "
                      "%(networks)s",
                      {"net_id": network.neutron_net_id,
                       "networks": networks})
            docker_networks = self.docker.networks(names=[network.name])
            LOG.debug("docker networks with name matching '%(name)s': "
                      "%(networks)s",
                      {"name": network.name,
                       "networks": docker_networks})
            if (networks and networks[0].network_id and
                    docker_networks and
                    networks[0].network_id == docker_networks[0]['Id']):
                LOG.debug("Network (%s) has already been created in docker",
                          network.name)
                return networks[0]
            else:
                # Probably, there are concurrent requests on creating the
                # network but the network is yet created in Docker.
                # We return False and let the caller retry.
                return False

        LOG.debug("Calling docker.create_network to create network %s, "
                  "ipam_options %s, options %s",
                  network.name, ipam_options, options)
        enable_ipv6 = bool(options.get('neutron.subnet.v6.uuid'))
        try:
            docker_network = self.docker.create_network(
                name=network.name,
                driver=CONF.network.driver_name,
                enable_ipv6=enable_ipv6,
                options=options,
                ipam=ipam_options)
        except Exception:
            with excutils.save_and_reraise_exception():
                network.destroy()

        network.network_id = docker_network['Id']
        network.save()
        return network

    def _get_subnet(self, subnets, ip_version):
        subnets = [s for s in subnets if s['ip_version'] == ip_version]
        if len(subnets) == 0:
            return None
        elif len(subnets) == 1:
            return subnets[0]
        else:
            raise exception.ZunException(_(
                "Multiple Neutron subnets exist with ip version %s") %
                ip_version)

    def remove_network(self, network):
        self.docker.remove_network(network.name)
        network.destroy()

    def process_networking_config(self, container, requested_network,
                                  host_config, container_kwargs, docker,
                                  security_group_ids):
        docker_net_name = requested_network['network']
        neutron_net_id = requested_network['network']
        addresses, port = self.neutron_api.create_or_update_port(
            container, neutron_net_id, requested_network, DEVICE_OWNER,
            security_group_ids, set_binding_host=True)
        container.addresses = {requested_network['network']: addresses}

        ipv4_address = None
        ipv6_address = None
        for address in addresses:
            if address['version'] == 4:
                ipv4_address = address['addr']
            if address['version'] == 6:
                ipv6_address = address['addr']

        endpoint_config = docker.create_endpoint_config(
            ipv4_address=ipv4_address, ipv6_address=ipv6_address)
        network_config = docker.create_networking_config({
            docker_net_name: endpoint_config})

        host_config['network_mode'] = docker_net_name
        container_kwargs['networking_config'] = network_config
        container_kwargs['mac_address'] = port['mac_address']

    def connect_container_to_network(self, container, requested_network,
                                     security_groups=None):
        """Connect container to the network

        This method will create a neutron port, retrieve the ip address(es)
        of the port, and pass them to docker.connect_container_to_network.
        """
        network_name = requested_network['network']
        container_id = container.container_id

        neutron_net_id = requested_network['network']
        addresses, original_port = self.neutron_api.create_or_update_port(
            container, neutron_net_id, requested_network, DEVICE_OWNER,
            security_groups)

        ipv4_address = None
        ipv6_address = None
        for address in addresses:
            if address['version'] == 4:
                ipv4_address = address['addr']
            if address['version'] == 6:
                ipv6_address = address['addr']

        kwargs = {}
        if ipv4_address:
            kwargs['ipv4_address'] = ipv4_address
        if ipv6_address:
            kwargs['ipv6_address'] = ipv6_address
        try:
            self.docker.connect_container_to_network(
                container_id, network_name, **kwargs)
        except exception.DockerError:
            with excutils.save_and_reraise_exception():
                self.do_port_cleanup(addresses, original_port)
        return addresses

    def do_port_cleanup(self, addresses, port):
        preserve_flag = addresses[0].get('preserve_on_delete')
        port_id = port.get('id')
        if preserve_flag:
            port_req_body = {'port': {'device_id': '', 'device_owner': ''}}
            port_req_body['port'][consts.BINDING_HOST_ID] = None
            port_req_body['port']['mac_address'] = port.get('mac_address')
            port_req_body['port'][consts.BINDING_PROFILE] = \
                port.get(consts.BINDING_PROFILE, {})

            try:
                # Requires admin creds to set port bindings
                self.neutron_api.update_port(port_id, port_req_body,
                                             admin=True)
            except exception.PortNotFound:
                LOG.debug('Unable to unbind port %s as it no longer '
                          'exists.', port_id)
            except Exception:
                LOG.exception("Unable to clear device ID for port '%s'",
                              port_id)
        else:
            try:
                self.neutron_api.delete_port(port_id)
            except exception.PortNotFound:
                LOG.debug('Unable to delete port %s as it no longer '
                          'exists.', port_id)

    def disconnect_container_from_network(self, container, neutron_network_id):
        network_name = neutron_network_id
        container_id = container.container_id

        addrs_list = []
        if container.addresses and neutron_network_id:
            addrs_list = container.addresses.get(neutron_network_id, [])

        self._disconnect_container_from_network(container_id, network_name,
                                                addrs_list)

    def _disconnect_container_from_network(self, container_id, network_name,
                                           addrs_list):
        neutron_ports = set()
        all_ports = set()
        for addr in addrs_list:
            all_ports.add(addr['port'])
            if not addr['preserve_on_delete']:
                port_id = addr['port']
                neutron_ports.add(port_id)

        try:
            if container_id:
                self.docker.disconnect_container_from_network(container_id,
                                                              network_name)
        finally:
            self.neutron_api.delete_or_unbind_ports(all_ports, neutron_ports)

    def add_security_groups_to_ports(self, container, security_group_ids):
        port_ids = set()
        for addrs_list in container.addresses.values():
            for addr in addrs_list:
                port_id = addr['port']
                port_ids.add(port_id)

        search_opts = {'tenant_id': self.context.project_id}
        neutron_ports = self.neutron_api.list_ports(
            **search_opts).get('ports', [])
        neutron_ports = [p for p in neutron_ports if p['id'] in port_ids]
        for port in neutron_ports:
            if 'security_groups' not in port:
                port['security_groups'] = []
            port['security_groups'].extend(security_group_ids)
            updated_port = {'security_groups': port['security_groups']}
            try:
                LOG.info("Adding security group %(security_group_ids)s "
                         "to port %(port_id)s",
                         {'security_group_ids': security_group_ids,
                          'port_id': port['id']})
                self.neutron_api.update_port(port['id'],
                                             {'port': updated_port},
                                             admin=True)
            except exceptions.NeutronClientException as e:
                exc_info = sys.exc_info()
                if e.status_code == 400:
                    raise exception.SecurityGroupCannotBeApplied(
                        str(e))
                else:
                    utils.reraise(*exc_info)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception("Neutron Error:")

    def remove_security_groups_from_ports(self, container, security_group_ids):
        port_ids = set()
        for addrs_list in container.addresses.values():
            for addr in addrs_list:
                port_id = addr['port']
                port_ids.add(port_id)

        search_opts = {'tenant_id': self.context.project_id}
        neutron_ports = self.neutron_api.list_ports(
            **search_opts).get('ports', [])
        neutron_ports = [p for p in neutron_ports if p['id'] in port_ids]
        for port in neutron_ports:
            port['security_groups'].remove(security_group_ids[0])
            updated_port = {'security_groups': port['security_groups']}
            try:
                LOG.info("Removing security group %(security_group_ids)s "
                         "from port %(port_id)s",
                         {'security_group_ids': security_group_ids,
                          'port_id': port['id']})
                self.neutron_api.update_port(port['id'],
                                             {'port': updated_port},
                                             admin=True)
            except exceptions.NeutronClientException as e:
                exc_info = sys.exc_info()
                if e.status_code == 400:
                    raise exception.SecurityGroupCannotBeRemoved(
                        str(e))
                else:
                    utils.reraise(*exc_info)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception("Neutron Error:")

    def on_container_started(self, container):
        pass

    def on_container_stopped(self, container):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pecan

from zun.api.controllers import base
from zun.api.controllers.v1.schemas import quota_classes as schema
from zun.api import validation
from zun.common import exception
from zun.common import policy
from zun.common import quota
from zun import objects

QUOTAS = quota.QUOTAS


class QuotaClassController(base.Controller):
    """Controller for QuotaClass"""

    def _get_quotas(self, context, quota_class):
        return QUOTAS.get_class_quotas(context, quota_class)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    @validation.validate_query_param(pecan.request, schema.query_param_update)
    @validation.validated(schema.query_param_update)
    def put(self, quota_class_name, **quota_classes_dict):
        context = pecan.request.context
        policy.enforce(context, 'quota_class:update',
                       action='quota_class:update')
        for key, value in quota_classes_dict.items():
            value = int(value)
            quota_class = objects.QuotaClass(
                context, class_name=quota_class_name,
                resource=key, hard_limit=value)
            try:
                quota_class.update(context)
            except exception.QuotaClassNotFound:
                quota_class.create(context)
        return self._get_quotas(context, quota_class_name)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get(self, quota_class_name):
        context = pecan.request.context
        policy.enforce(context, 'quota_class:get',
                       action='quota_class:get')
        return self._get_quotas(context, quota_class_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shlex

from oslo_log import log as logging

from zun.common import consts
from zun.common import context as zun_context
from zun.common import exception
from zun.common import utils
import zun.conf
from zun.network import network
from zun.network import neutron
from zun.network import os_vif_util
from zun import objects

CONF = zun.conf.CONF

LOG = logging.getLogger(__name__)

DEVICE_OWNER = 'compute:zun'

ZUN_INIT_NETWORK_READY_PATH = '/zun-init-network-ready'
ZUN_INIT = ('while ! [ -f %(file)s ] ; '
            'do sleep 1 ; done ; rm -f %(file)s ; exec '
            ) % {'file': ZUN_INIT_NETWORK_READY_PATH}


ZUN_CNI_CONF = ""
with open(CONF.cni_daemon.zun_cni_config_file, 'r') as f:
    ZUN_CNI_CONF = f.read()

PATH = os.environ['PATH']
if "CNI_PATH" in os.environ:
    PATH = os.environ['CNI_PATH'] + ':' + PATH

ZUN_CNI_BIN = "zun-cni"
ZUN_CNI_ADD_CMD = "ADD"
ZUN_CNI_DEL_CMD = "DEL"


class ZunCNI(network.Network):
    def init(self, context, docker_api):
        self.docker = docker_api
        self.neutron_api = neutron.NeutronAPI(context)
        self.context = context

    def get_or_create_network(self, *args, **kwargs):
        pass

    def create_network(self, name, neutron_net_id):
        pass

    def remove_network(self, network):
        pass

    def connect_container_to_network(self, container, requested_network,
                                     security_groups=None):
        # TODO(hongbin): implement this in zun-cni-daemon
        pass

    def disconnect_container_from_network(self, container, neutron_network_id):
        addrs_list = []
        if container.addresses and neutron_network_id:
            addrs_list = container.addresses.get(neutron_network_id, [])

        neutron_ports = set()
        all_ports = set()
        for addr in addrs_list:
            all_ports.add(addr['port'])
            if not addr['preserve_on_delete']:
                port_id = addr['port']
                neutron_ports.add(port_id)

        self.neutron_api.delete_or_unbind_ports(all_ports, neutron_ports)

    def process_networking_config(self, container, requested_network,
                                  host_config, container_kwargs, docker,
                                  security_group_ids=None):
        network_id = requested_network['network']
        addresses, port = self.neutron_api.create_or_update_port(
            container, network_id, requested_network, consts.DEVICE_OWNER_ZUN,
            security_group_ids, set_binding_host=True)
        container.addresses = {network_id: addresses}

        admin_neutron_api = neutron.NeutronAPI(zun_context.get_admin_context())
        network = admin_neutron_api.show_network(port['network_id'])['network']
        subnets = {}
        for fixed_ip in port['fixed_ips']:
            subnet_id = fixed_ip['subnet_id']
            subnets[subnet_id] = \
                admin_neutron_api.show_subnet(subnet_id)['subnet']
        vif_plugin = port.get('binding:vif_type')
        vif_obj = os_vif_util.neutron_to_osvif_vif(vif_plugin, port, network,
                                                   subnets)
        state = objects.vif.VIFState(default_vif=vif_obj)
        state_dict = state.obj_to_primitive()
        container.cni_metadata = {consts.CNI_METADATA_VIF: state_dict}
        container.save(self.context)

        host_config['network_mode'] = 'none'
        container_kwargs['mac_address'] = port['mac_address']

        # We manipulate entrypoint and command parameters in here.
        token = (container.entrypoint or []) + (container.command or [])
        new_command = ZUN_INIT + ' '.join(shlex.quote(t) for t in token)
        new_entrypoint = ['/bin/sh', '-c']
        container_kwargs['entrypoint'] = new_entrypoint
        container_kwargs['command'] = [new_command]

    def _get_env_variables(self, container, pid, cmd):
        return {
            'PATH': PATH,
            'CNI_COMMAND': cmd,
            'CNI_CONTAINERID': str(container.container_id),
            'CNI_NETNS': '/proc/%s/ns/net' % pid,
            'CNI_ARGS': "K8S_POD_NAME=%s;ZUN_CONTAINER_TYPE=CONTAINER" %
                        container.uuid,
            'CNI_IFNAME': 'eth0',
        }

    def _add(self, container, pid):
        env_variables = self._get_env_variables(container, pid,
                                                ZUN_CNI_ADD_CMD)
        utils.execute(ZUN_CNI_BIN,
                      process_input=ZUN_CNI_CONF,
                      env_variables=env_variables)

    def _delete(self, container, pid):
        env_variables = self._get_env_variables(container, pid,
                                                ZUN_CNI_DEL_CMD)
        utils.execute(ZUN_CNI_BIN,
                      process_input=ZUN_CNI_CONF,
                      env_variables=env_variables)

    def _exec_command_in_container(self, container, cmd):
        exec_id = self.docker.exec_create(container.container_id, cmd)['Id']
        output = self.docker.exec_start(exec_id)
        inspect_res = self.docker.exec_inspect(exec_id)
        return inspect_res['ExitCode'], output

    def on_container_started(self, container):
        response = self.docker.inspect_container(container.container_id)
        state = response.get('State')
        if type(state) is dict and state.get('Pid'):
            pid = state['Pid']
        else:
            pid = None
        # no change
        cni_metadata = container.cni_metadata
        if cni_metadata.get(consts.CNI_METADATA_PID) == pid:
            return

        # remove container from old network
        old_pid = cni_metadata.pop(consts.CNI_METADATA_PID, None)
        if old_pid:
            self._delete(container, old_pid)
            container.cni_metadata = cni_metadata
            container.save(self.context)
        # add container to network
        self._add(container, pid)
        cni_metadata[consts.CNI_METADATA_PID] = pid
        container.cni_metadata = cni_metadata
        container.save(self.context)

        # notify the container that network is setup
        cmd = ['touch', ZUN_INIT_NETWORK_READY_PATH]
        exit_code, output = self._exec_command_in_container(container, cmd)
        if exit_code != 0:
            raise exception.ZunException('Execute command %(cmd)s failed, '
                                         'output is: %(output)s'
                                         % {'cmd': ' '.join(cmd),
                                            'output': output})

    def on_container_stopped(self, container):
        cni_metadata = container.cni_metadata
        pid = cni_metadata.pop(consts.CNI_METADATA_PID, None)
        if pid:
            self._delete(container, pid)
            container.cni_metadata = cni_metadata
            container.save(self.context)



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add host to image

Revision ID: a019998b09b5
Revises: a9c9fb54274a
Create Date: 2018-08-17 13:49:11.470002

"""

# revision identifiers, used by Alembic.
revision = 'a019998b09b5'
down_revision = 'a9c9fb54274a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('image',
                  sa.Column('host', sa.String(length=255), nullable=True))
    op.drop_constraint(constraint_name='uniq_image0repotag',
                       table_name='image', type_='unique')
    op.create_unique_constraint(constraint_name='uniq_image0repotaghost',
                                table_name='image',
                                columns=['repo', 'tag', 'host'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg


neutron_group = cfg.OptGroup(name='neutron_client',
                             title='Options for the Neutron client')

common_security_opts = [
    cfg.StrOpt('ca_file',
               help='Optional CA cert file to use in SSL connections.'),
    cfg.StrOpt('cert_file',
               help='Optional PEM-formatted certificate chain file.'),
    cfg.StrOpt('key_file',
               help='Optional PEM-formatted file that contains the '
                    'private key.'),
    cfg.BoolOpt('insecure',
                default=False,
                help="If set, then the server's certificate will not "
                     "be verified.")]

neutron_client_opts = [
    cfg.StrOpt('region_name',
               help='Region in Identity service catalog to use for '
                    'communication with the OpenStack service.'),
    cfg.StrOpt('endpoint_type',
               default='publicURL',
               help='Type of endpoint in Identity service catalog to use '
                    'for communication with the OpenStack service.')]


ALL_OPTS = (neutron_client_opts + common_security_opts)


def register_opts(conf):
    conf.register_group(neutron_group)
    conf.register_opts(ALL_OPTS, group=neutron_group)


def list_opts():
    return {neutron_group: ALL_OPTS}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from zun.common import context
from zun import objects
from zun.scheduler.filters import cpu_filter
from zun.tests import base
from zun.tests.unit.scheduler import fakes


class TestCPUFilter(base.TestCase):

    def setUp(self):
        super(TestCPUFilter, self).setUp()
        self.context = context.RequestContext('fake_user', 'fake_project')

    def test_cpu_filter_pass(self):
        self.filt_cls = cpu_filter.CPUFilter()
        container = objects.Container(self.context)
        container.cpu = 5.0
        host = fakes.FakeHostState('testhost')
        host.cpus = 8
        host.cpu_used = 0.0
        extra_spec = {}
        self.assertTrue(self.filt_cls.host_passes(host, container, extra_spec))

    def test_cpu_filter_fail(self):
        self.filt_cls = cpu_filter.CPUFilter()
        container = objects.Container(self.context)
        container.cpu = 8.0
        host = fakes.FakeHostState('testhost')
        host.cpus = 5
        host.cpu_used = 2.0
        extra_spec = {}
        self.assertFalse(self.filt_cls.host_passes(host, container,
                                                   extra_spec))
